
#' Calculate cluster statistics
#' 
#' Calculate counts and percent of total for each cluster and source file
#' 
#' @name clusterCounts
#' @aliases clusterFract
#' 
#' @param tmap     tmap object containing cluster info, as returned by cluster.tmap
#' 
#' @return 
#' A matrix with rows determined by \code{tmap$sourceFile}, and columns 
#' determined by \code{tmap$clustInfo$cluster}. \code{clusterCounts} returns a matrix of cell 
#' counts, while \code{clusterFract} returns a count matrix normalized by the total number
#' of cells in each row
#' 
#' 
#' @export
clusterCounts <- function(tmap){
  ##first, check inputs
  if(class(tmap)!="tmap"){
    stop("tmap must be a valid 'tmap' object")
  }
  if(is.null(tmap$clustInfo)){
    stop("cluster information not found")
  }
  
  ##get cluster counts
  cts = table(tmap$sourceFile, tmap$clustInfo$cluster)
  matrix(cts, nrow=nrow(cts), ncol=ncol(cts), dimnames = dimnames(cts))
}
#' @rdname clusterCounts
#' @export
clusterFract <- function(tmap){
  cts = clusterCounts(tmap)
  cts/rowSums(cts)
}


#' Compare cluster membership
#' 
#' @description Methods for calculating significant differences in cluster membership 
#' between groups of samples.
#' 
#' @param tmap        an object of class "\code{tmap}" containing clustering information.
#' @param grp.info    data frame with two columns, OR a named vector; a description of 
#'                    the sample grouping in tmap. The first column (or the names of the 
#'                    vector) is a list of filenames that match the values in 
#'                    \code{tmap$sourceFile}, and the second column (or values in the 
#'                    vector) is the group membership of that file.
#' @param contrast    character string describing the contrast of interest. Must be of the
#'                    form "GrpB-GrpA", where groups A and B are the group labels to be 
#'                    compared. 
#' @param parametric  boolean; should a parametric test be used? 
#' @param formula     an object of class "\code{\link[stats]{formula}}" (or one that can 
#'                    be coerced to that class): a symbolic description of which fields 
#'                    to use to group the data in tmap.
#' @param clust.fract Matrix (or object that can be coerced to a matrix) containing
#'                    cluster membership fractions, as returned by 
#'                    \code{\link{clusterFract}}.
#' @param alternative,var.equal    parameter to pass to t.test or wilcox.test
#'                 
#' @return A data frame with the following information:
#' \describe{
#'  \item{cluster}{The cluster ID}
#'  \item{statistic}{The t-statistic (if \code{parametric=TRUE}) or W-statistic (if 
#'                        \code{FALSE})}
#'  \item{p.value}{The p-value}
#'  \item{fdr}{The false discovery rate}
#' }
#' @export             
compareGroups <- function(x, ...) UseMethod("compareGroups")
#' @rdname compareGroups
#' @export
compareGroups.tmap <- function(tmap, grp.info, contrast, parametric=FALSE, ...){
  clust.fract = clusterFract(tmap)
  compareGroups.default(clust.fract, grp.info, contrast, parametric=parametric,...)
}
#' @rdname compareGroups
#' @export
compareGroups.formula <- function(formula, data, contrast, tmap=NULL, clust.fract=NULL,...){
  
  ##get cluster fractions (and check inputs)
  if(!is.null(tmap)){
    if(!is.null(clust.fract)){
      warning("Both tmap and clust.fract provided. Only using information in clust.fract")
    }
    clust.fract = clusterFract(tmap)
  }
  if(is.null(clust.fract)){
    stop("Either a valid tmap object or a matrix containing cluster fractions must be provided")
  }
  
  ##get groups 
  data = as.data.frame(data)
  if(!is.data.frame(data)){stop("could not coerce data to data.frame")}
  
  mf <- model.frame(formula, data=data)
  
  grp.info = data.frame(files = mf[,1],
                        grps = do.call(function(...){paste(...,sep="_")}, mf[,-1,drop=F])
  )
  
  ##compare groups
  compareGroups.default(clust.fract, grp.info, contrast=contrast, ...)
}
#' @rdname compareGroups
#' @export
compareGroups.default <- function(clust.fract, grp.info, contrast, parametric=FALSE,
                                  alternative = c("two.sided", "less", "greater"),
                                  var.equal=FALSE){
  
  ##check inputs 
  alternative = match.arg(alternative)
  
  ##check that clust.fract is a matrix
  if(!is.matrix(clust.fract)){
    clust.fract = as.matrix(clust.fract)
    if(!is.matrix(clust.fract)){
      stop("cannot coerce clust.fract to matrix")
    }
  }
  
  ##coerce grp.info to two-column data frame
  if(!is.data.frame(grp.info)){
    grp.info = as.data.frame(grp.info)
    if(ncol(grp.info)==1 && !is.null(rownames(grp.info))){
      grp.info = as.data.frame(cbind(sample.name=rownames(grp.info), grp.info))
    }
  }
  if(!is.data.frame(grp.info) || ncol(grp.info) != 2){
    stop("cannot coerce grp.info to 2-column data frame.")
  }
  
  ##check that sample names match rownames in clust.fract
  if(!any(grp.info[,1] %in% rownames(clust.fract))){
    stop("sample names not found in clust.fract")
  }
  grp.info = grp.info[grp.info[,1] %in% rownames(clust.fract),]
  
  ##check contrast format
  if(!is.character(contrast)){stop("Contrast must be a character vector of length 1.")}
  if(length(contrast)!=1){
    warning("Multiple contrasts provided. Using first contrast only.")
    contrast = contrast[1]
  }
  
  ##parse the contrast
  grps = strsplit(contrast,"-")[[1]]
  grps = sub("\\s","",grps)          ##remove whitespace
  if(length(grps)!=2){stop("Only contrasts of the form 'GrpA-GrpB' are allowed.")}
  grp.1 = grp.info[,2]==grps[1]  ##PostTreatment
  grp.2 = grp.info[,2]==grps[2]  ##Baseline
  
  if(sum(grp.1)==0 | sum(grp.2)==0){stop("Contrast groups do not match group labels")}
  
  ##get group filenames
  grp.1 = as.character(grp.info[grp.1,1])
  grp.2 = as.character(grp.info[grp.2,1])
  
  g1.mat = clust.fract[grp.1,]
  g2.mat = clust.fract[grp.2,]
  
  ##calculate statistics
  stats = sapply(1:ncol(g1.mat), function(i){
    if(parametric){
      x=t.test(g1.mat[,i], g2.mat[,i], alternative = alternative, var.equal = var.equal)
    }else{
      x=wilcox.test(g1.mat[,i], g2.mat[,i], alternative=alternative, var.equal=var.equal, 
                  exact=FALSE)
    }
    c(statistic=x$statistic,p.value = x$p.value)
  })
  stats = data.frame(cluster=colnames(clust.fract),t(stats))
  stats$fdr = p.adjust(stats$p.value, method = "fdr")
  
  ##format the stats matrix
  names(stats) = c("cluster","statistic","p.value","fdr")
  rownames(stats) = NULL
  stats[,3:4] = signif(stats[,3:4], 3)
  
  o = order(stats$fdr,stats$p.value, stats$statistic)
  
  stats[o,]
}