
#' cluster.tmap
#' 
#' @name cluster.tmap
#' @description cluster a high density 2-dimensional tSNE map using 'adjacency clustering'
#' 
#' @param tmap         tmap object containing a 2-dimensional tSNE.map.
#' @param nclust       desired number of clusters 
#' @param minSize      The minimum number of cells to define a cluster. Any clusters with
#'                     a number of cells smaller than minSize will not be counted towards 
#'                     minSize, and will be annotated with \code{NA} for the cluster group, 
#'                     indicating they don't belong to any cluster.
#' @param xbins        number of bins partitioning the range of the x dimension.
#' @param weighted     logical; if true, adjacency clustering is performed using weighted
#'                     distances (see adj_clust for complete details)
#' @param markerWt     Define how to weight individual markers. Can either be a numeric
#'                     vector of the same length as the columns in tmap$markerVals, a
#'                     named vector indicating the weights of specific markers, or
#'                     "SD" to calculate weights relative to the standard 
#'                     deviation of each parameter within each bin (see details).
#' @param mapWt        value between 0 and 1 indicating the contribution of distances 
#'                     between points in the 2-dimensional map to the total distance.
#'                     A value of 1 implies that distances are calculated based on only 
#'                     the 2-dimensional map, whereas a value of 0 means only the original
#'                     n-dimensional map is used. 
#' @param findEdges    boolean; if true, the cluster edges will be determined and added
#'                     to the tmap object. Adds significant processing time.
#'                     
#' @details         This function uses the 'adjacency' hierarchical clustering algorithm
#'                  to cluster high-density data in a 2-dimensional map. 
#'                  
#'                  TODO: talk about markerWt
#'                  
#' @export
cluster.tmap <- function(tmap, nclust=NULL, minSize=40,
                         method=c("adjClust","densVM","citrus","hierarchical","kmeans"), 
                         hclust.meth = "complete",
                         xbins=150, weighted=F,
                         markerWt="none", mapWt = 0, findEdges=T){  
  ##check parameters
  method = match.arg(method)
  if(method=="adjClust" && is.null(nclust)){stop("Must specify number of clusters if method is 'adjClust'")}
  if(mapWt<0 || mapWt>1){stop("mapWt must be between 0 and 1")}
  
  ################ densVM #################
  if(method == "densVM"){
    stop("This method has not yet been fully implemented")
#     ##find bandwidth
#     toUse = 1:nrow(tmap)
#     if(nrow(tmap)>50000){toUse=sample(toUse, 50000)}
#     ydata = tmap$tSNE.map[toUse, ]
#     
#     ##-----------
#     ##code taken from cytofkit package
#     y_range <- min(c(diff(range(ydata[, 1])), diff(range(ydata[, 2]))))
#     
#     ## Search density peaks of different kernel bandwidth(gamma or sig_opt)
#     num_band <- 20
#     sig_tol_range <- seq(y_range/100, y_range/10, length.out = num_band)
#     kernel_min = ___
#     kernel_max = ___
#     
#     cat("Testing kernel bandwidth for ", num_band, "points in the range min=", 
#         kernel_min, " to max=", kernel_max, "\n")
#     
#     cat("This will take a while...\n")
#     peak_list <- mapply(peakFind, sig_tol = sig_tol_range, MoreArgs = list(ydata = ydata), 
#                         SIMPLIFY = FALSE)
#     
#     Npeaks <- sapply(peak_list, function(x) { x$Npeaks })
#     peakdata <- data.frame(numpeaks = Npeaks, sig_range = sig_tol_range)
#     
#     ## Locate plateau(knee)
#     sig_opt = which(-diff(Npeaks)<=1)[1]
#     ## if none are <= 1, just use the lowest
#     if(is.na(sig_opt)){sig_opt = which.min(-diff(Npeaks))}
#     
#     ##now run peakFind on all data
#     opt_densityIMG <- peakFind(ydata, sig_opt)
#     
#     ##and assign clusters
#     clusters <- assign_pop(opt_densityIMG, tmap$tSNE.map,tmap$markerVals)
#     
#     tmap$cluster = as.numeric(clusters[,3])
#     if(!is.null(tmap$binInfo)) tmap$binInfo = NULL
    
  ################ hierarchical clustering #################
  }else if(method == "adjClust" || method == "citrus" || method == "hierarchical"){
    ##get bin info
    if(!is.null(tmap$binInfo) && tmap$binInfo$hexbin@xbins == xbins){
      binInfo = tmap$binInfo
      cat("using existing bin information\n")
    }else{
      binInfo = hexSplit.tmap(tmap,xbins = xbins)
      tmap$binInfo = binInfo
    }  
    
    d = calcBinDist(tmap, markerWt=markerWt, mapWt=mapWt)
    
    ##weight distances by number of cells in smaller bin
#     cts = binInfo$hexbin@count
#     pmin.ct = unlist(lapply(1:(length(cts)-1), function(i){
#       j = (i+1):length(cts)
#       pmin(cts[rep(i,length(j))], cts[j])
#     }))
#     ctWt = 1/(1+log2(pmin.ct))
#     
#     ##apply to dist
#     d = d*ctWt
    
    if(method=="adjClust"){
      ##create indicator vector for adjacency
      n=attr(d, "Size")
      isadj = lapply(1:n,function(i){
        nbrs = binInfo$neighbor.map[[i]]
        ii = rep(i, length(nbrs)); jj = nbrs
        ii[nbrs<i] = nbrs[nbrs<i]; jj[nbrs<i] = i
        distdex(ii,jj,n)
      })
      adj = rep(F, length(d))
      adj[unlist(isadj)] = T
      
      ##run adjacency clustering
      hc = hclust_adj(d,adj, weighted=weighted, method=hclust.meth)
    }
    ##split into nclust megaclusters
    clustMap = cutree(hc, k=nclust)
    cluster = clustMap[binInfo$bin]
    
    ##find the actual number of clusters using (sorta) binary search
    nclust.act = nclust
    while( (clustCount = sum(table(cluster)>=minSize)) != nclust ){
      diff = nclust - clustCount
      if(diff>0){nclust.act = nclust.act + 2*diff}
      else{nclust.act = nclust.act + diff}
      
      clustMap = cutree(hc, k=nclust.act)
      cluster = clustMap[binInfo$clust]
    }
    
    ##now remove small clusters
    tooSmall = which(table(cluster)<minSize)
    if(length(tooSmall)>0){
      mapMap = rep(NA, nclust.act); mapMap[-tooSmall] = 1:nclust
      clustMap = mapMap[clustMap]
      cluster = mapMap[cluster]
    }
    
    ##find adjacent clusters
    adj = findAdjacentClusters(clustMap, binInfo$neighbor.map)
    
    #store everything in tmap
    clustInfo = list(nclust=nclust,
                     cluster=cluster, 
                     binMap = clustMap,
                     adjacency = adj
                     )
    tmap$clustInfo = clustInfo
    
    ##find the cluster edges
    if(findEdges){
      edges = findClustEdges(tmap)
      clustInfo$clustEdges = edges
      ##store in tmap
      tmap$clustInfo = clustInfo
    }
    
    return(tmap)
  }
}

#' Calculate distance between bins using various weighting schema.
#' @param tmap         the tmap object, with binInfo data 
#' @param markerWt     Define how to weight individual markers. Can either be a numeric
#'                     vector of the same length as the columns in tmap$markerVals, a
#'                     named vector indicating the weights of specific markers, or
#'                     "SD" (default) to calculate weights relative to the standard 
#'                     deviation of each parameter within each bin (see details).
#' @param mapWt        value between 0 and 1 indicating the contribution of distances 
#'                     between points in the 2-dimensional map to the total distance.
#'                     A value of 1 implies that distances are calculated based on only 
#'                     the 2-dimensional map, whereas a value of 0 means only the original
#'                     n-dimensional map is used. 
calcBinDist = function(tmap, markerWt="none", mapWt = 0){
  binInfo = tmap$binInfo
  bin = binInfo$clust
  
  ##get average of markers for each bin
  bin.means = sapply(split(1:length(bin), bin), function(i){
    apply(tmap$markerVals[i,,drop=F], 2, mean)
  })
  ##add in the map info
  bin.means = cbind(t(bin.means),tmap$binInfo$means)
  
  ##########-- calcluate marker weights --##########
  
  if(!is.null(markerWt) && 
     ( length(markerWt)!=1 || (!is.na(markerWt) && markerWt!="none") ) ){
    if(length(markerWt)==1 && markerWt=="SD"){
      
      ##calculate weights for each marker based on standard deviation
      bin.sds = sapply(split(1:length(bin), bin), function(i){
        apply(tmap$markerVals[i,,drop=F], 2, sd)
      })
      wts = 1/(1+bin.sds)
      wts[is.na(wts)] = 0.5
      wts = t(t(wts) / colSums(wts) * nrow(wts))        
    }else if(is.numeric(markerWt)){
      ##use vector of marker weights 
      ##if named, add any values that aren't defined
      if(!is.null(names(markerWt))){
        notDef = colnames(tmap$markerVals)[!(colnames(tmap$markerVals) %in% names(markerWt))]
        markerWt[notDef] = 1
        markerWt = markerWt[colnames(tmap$markerVals)]
      }
      if(length(markerWt)!=ncol(tmap$markerVals)){
        stop("if names are not provided, length of markerWt must be equal to ncol(tmap$markerVals)")
      }
      wts = matrix(markerWt, ncol(tmap$markerVals), nrow(bin.means))
    }
    ##if markerWt is NULL, NA, or "none", set all weights to 1
  }else{wts = matrix(1, ncol(tmap$markerVals) ,nrow(bin.means))}
  
  ##add in the map weight
  wts = cbind(t(wts) * (1-mapWt), matrix(mapWt, ncol(wts), ncol(tmap$binInfo$means)) )
  
  
  ##calculate distances
  d = dist_wtd(bin.means, wts)
  return(d)
}

#' Fortran implementation of hclust_adj
#' 
#' @name hclust_adj
#' @description     Fortran implementation of hclust_adj
#' 
#' @param d         a dissimilarity structure, as defined by 'dist'
#' @param adj       logical vector of same length as d indicating the adjacency of each 
#'                  pair of points.
#' @param weighted  logical; if true, adjacency clustering is performed using weighted
#'                  distances (see details)
#' 
#' @details         There are two forms of the adjacency clustering algorithm; weighted
#'                  and unweighted. For unweighted adjacency clustering, distance between
#'                  bins is calculated using standard euclidean distance, and hierarchical
#'                  clustering groups the most similar adjacent clusters at every step. If
#'                  weighted=T is specified, then a weighting parameter is included for 
#'                  smaller clusters such that distance to any cluster with more than 2 
#'                  members is 1.2 times higher than to a cluster with only 1 member.
#'                  
#' @return          An object of class hclust which describes the tree produced by 
#'                  the clustering process. See hclust for full details          
#' @export
hclust_adj <- function(d, adj, method="complete", weighted=F, members=NULL){
  ##match method
  METHODS <- c("ward.D", "single", "complete", "average", "mcquitty", 
               "median", "centroid", "ward.D2")
  if (method == "ward") {
    message("The \"ward\" method has been renamed to \"ward.D\"; note new \"ward.D2\"")
    method <- "ward.D"
  }
  i.meth <- pmatch(method, METHODS)
  if (is.na(i.meth)) 
    stop("invalid clustering method", paste("", method))
  if (i.meth == -1) 
    stop("ambiguous clustering method", paste("", method))
  
  ##get params
  n <- as.integer(attr(d, "Size"))
  if(is.null(n))
    stop("invalid dissimilarities")
  if(is.na(n) || n > 65536L) stop("size cannot be NA nor exceed 65536")
  if(n < 2)
    stop("must have n >= 2 objects to cluster")
  len <- as.integer(n*(n-1)/2)
  if(length(d) != len)
    (if (length(d) < len) stop else warning
    )("dissimilarities of improper length")
  if(length(adj) != len)
    (if (length(adj) != length(d)) stop)
     
  if(is.null(members)) members <- rep(1, n)
  else if(length(members) != n) stop("invalid length of members")
      
   storage.mode(d) <- "double"
   hcl <- .Fortran("adjclust",
                   n = n,
                   len = len,
                   iopt = as.integer(i.meth),
                   ia = integer(n),
                   ib = integer(n),
                   crit = double(n),
                   membr = as.double(members),
                   nn = integer(n),
                   disnn = double(n),
                   flag = logical(n),
                   diss = d,
                   adj = adj,
                   wtd = weighted)
 
 ## 2nd step: interpret the information that we now have
 ## as merge, height, and order lists.
 
 hcass <- .Fortran("hcass2",
                   n = n, # checked above.
                   ia = hcl$ia,
                   ib = hcl$ib,
                   order = integer(n),
                   iia = integer(n),
                   iib = integer(n))
 
 structure(list(merge = cbind(hcass$iia[1L:(n-1)], hcass$iib[1L:(n-1)]),
                height = hcl$crit[1L:(n-1)],
                order = hcass$order,
                labels = attr(d, "Labels"),
                method = "adjClust",
                call = match.call(),
                dist.method = attr(d, "method")),
           class = "hclust")
}

#' hexSplit_tSNE
#' 
#' @name hexSplit
#' @description split 2-dimensional map into hexagonal bins and define neighbors
#' 
#' @param tmap     tmap object containing a 2-dimensional tSNE.map
#' @param xbins    number of bins along x-axis to split map into.
#' @param min.nbr  minimum number of neighbors for each bin. Cannot exceed 6
#' @export
hexSplit.tmap <- function(tmap, xbins=75, min.nbr=3){
  ##check params
  tSNE.map = tmap$tSNE.map
  if(is.null(tSNE.map)){stop("tSNE.map not found")}
  
  if(min.nbr>6){stop("min.nbr cannot exceed 6")}
  
  
  ##instead of k-means, just split cells up into equally-sized hexes
  hexes = hexbin(tSNE.map[,1],tSNE.map[,2], xbins = xbins,IDs=T)
  
  ##redefine cells so there's a buffer on the edges
  hexes = addEdgeBuffer(hexes)
  
  cell <- hexes@cell - 1
  ncol=hexes@dimen[2]
  nr <- cell%/%ncol + 1
  
  means = cbind(hexes@xcm, hexes@ycm)
  tSNE.dist = dist(means)
  
  ##neighbor definitions
  nVec.1 = c(-1, ncol, ncol+1, 1, -ncol+1,-ncol)
  nVec.2 = c(-1, ncol-1, ncol, 1, -ncol,-ncol-1)
  
  ##find neighbors
  neighbors = lapply(1:length(cell),function(i){
    n = cell[i]+nVec.1
    if(nr[i]%%2==1){n=cell[i]+nVec.2}
    n = n[n %in% cell]
    if(length(n) < min.nbr){
      n.i = match(n,cell)
      d = tSNE.dist[distdex(i,-n.i,nrow(means))]
      cl = cell[-c(i,n.i)]
      while(length(n) < min.nbr){
        m.i = which.min(d)
        n = c(n,cl[m.i])
        d = d[-m.i]; cl = cl[-m.i]
      }
    }
    
    ###convert neighbor info back to indexes
    match(n,cell)
  })
  
  clust.all = match(hexes@cID, hexes@cell)
  
  return(list(nclust=hexes@ncells, clust=clust.all, neighbor.map=neighbors, 
              means=means, hexbin=hexes))
}

#' Combine multiple clusters together into a single cluster. 
#' @param tmap       the tmap object
#' @param combine    numeric vector; the clusters to combine together. All clusters in the
#'                   vector will be combined into a single cluster. combine can also be a 
#'                   list, in which case each list entry will be combined into a separate 
#'                   cluster.
#' @return           the updated tmap object.
#' 
#' @export
combineClusters = function(tmap, combine, rename=T){
  ##standard checks
  if(is.null(tmap$clustInfo)){stop("Clustering information not found")}
  
  if(!is.list(combine)){
    if(!is.numeric(combine)){stop("combine must be a numeric vector or list of vectors")}
    combine = list(combine)
  }
  if(!all(sapply(combine, is.numeric))){stop("All entris in cluster must be numeric")}
  
  ##generate the list that we're gonna shift everything to; 
  ## by default everything stays the same
  to = 1:(tmap$clustInfo$nclust)
  
  ##reassign everything in x to the first entry in x
  for(x in combine){to[x] = x[1]}
  
  ##if we want to rename clusters, do so here
  if(rename){
    old = unique(to)
    nclust = length(old)
    new = setNames(1:nclust, old)
    to = new[as.character(to)]
    names(to) = NULL
    
    ##save the new nclust 
    tmap$clustInfo$nclust = nclust
  }
  
  ##reassign everything in tmap
  tmap$clustInfo$cluster = to[tmap$clustInfo$cluster]
  tmap$clustInfo$binMap = to[tmap$clustInfo$binMap]
  
  
  ##recalc the clusterEdge information. It's no longer correct.
  tmap$clustInfo$clustEdges = findClustEdges(tmap)
  tmap
}                   

#' Generate a report on individual clusters describing the parameters that define them.
#' @param tmap       the tmap object
#' @param alpha      cutoff defining which markers to consider significantly different
#' @export
clusterReport = function(tmap, alpha=0.05, as.text=TRUE){
  es = markerES(tmap)
  
  es.all = es$es.vs.all
  es.nbr = es$es.vs.nbr
  
  fullCutoff = quantile(es.all, c(alpha/2, 1-alpha/2))
  nbrCutoff = quantile(unlist(es.nbr), c(alpha/2, 1-alpha/2))
  
  nbrSigs = lapply(es.nbr, function(esMat){
    if(ncol(esMat)==0){return(NULL)}
    sigs = lapply(1:ncol(esMat), function(j){
      x = esMat[,j]<nbrCutoff[1] | esMat[,j]>nbrCutoff[2]
      y = esMat[x,j]
      names(y) = rownames(esMat)[x]
      y
    })
    names(sigs) = colnames(esMat)
    sigs
  })
  
  fullSig = lapply(1:ncol(es.all), function(j){
    x = es.all[,j]<fullCutoff[1] | es.all[,j]>fullCutoff[2]
    es.all[x,j]
  })
  
  if(!as.text){return(list(nbr=nbrSigs, full=fullSig))}
  
  ##convert into text report
  sapply(1:length(nbrSigs), function(cl){
    ##vs neighbors
    nbr = nbrSigs[[cl]]
    nbrReport = sapply(names(nbr), function(n){
      sig = nbr[[n]]
      str = paste0("Compared to ",n,":")
      if(length(sig)<0){str = paste(str, "No significant differences")}
      if(any(sig<0)){
        str = paste0(str, " Low in ", paste(names(sig)[sig<0],collapse = ", "),";")
      }
      if(any(sig>0)){
        str = paste0(str, " High in ", paste(names(sig)[sig>0],collapse = ", "),";")
      }
      str
    })
    
    names(nbrReport) = NULL
    
    sig = fullSig[[cl]]
    totReport=NULL
    if(length(sig)>0){
      str = paste0("Compared to total map:")
      if(any(sig<0)){
        str = paste0(str, " Low in ", paste(names(sig)[sig<0],collapse = ", "),";")
      }
      if(any(sig>0)){
        str = paste0(str, " High in ", paste(names(sig)[sig>0],collapse = ", "),";")
      }
      totReport = str
    }
    
    return(c(nbrReport, totReport))
  })
  
}

#' Calculate differences (effect sizes) in markers between clusters
#' 
#' @return list with two elements: 
#'  - es.vs.all: matrix with comparison of each cluster with the whole map
#'  - es.vs.nbr: list of matrices: comparison of each cluster with its immediate neighbors
markerES <- function(tmap){
  ##pull out stuff from tmap
  if(is.null(tmap$clustInfo)){stop("cluster information not found")}
  clustInfo = tmap$clustInfo
  markerVals = tmap$markerVals
  nclust = clustInfo$nclust
  
  ##calculate effect sizes for each marker between a cluster and its neighbors
  es = lapply(1:nclust, function(i){
    inI = which(clustInfo$cluster == i)
    esMat = NULL
    ##vs the pairs
    if(length(clustInfo$adjacency[[i]])!=0){
      for(j in clustInfo$adjacency[[i]]){
        inJ = which(clustInfo$cluster == j)
        es = apply(markerVals, 2, function(x){esFormula(x[inI], x[inJ])})
        esMat = cbind(esMat, es)
      }
      colnames(esMat) = clustInfo$adjacency[[i]]
    }
    
    ##vs everything
    es = apply(markerVals, 2, function(x){esFormula(x[inI], x[-inI])})
    cbind(vsAll=es, esMat)
  })
  es.vs.all = sapply(es, function(x){x[,1]})
  es.vs.nbr = sapply(es, function(x){x[,-1, drop=F]})
  return(list(es.vs.all=es.vs.all, es.vs.nbr=es.vs.nbr))
}

esFormula <- function(x1, x2){
  (mean(x1) - mean(x2)) / ( var(x1) + var(x2) )
}


#' Method for finding cluster edges
#' @param tmap
#' 
findClustEdges <- function(tmap){
  ##make sure clustering has been run
  if(is.null(tmap$binInfo) || is.null(tmap$clustInfo)){
    stop("Cluster information not found")
  }
  
  ##pull out hex info
  hexes = tmap$binInfo$hexbin
  nrow = hexes@dimen[1]
  ncol = hexes@dimen[2]
  cell = hexes@cell
  
  ##pull out cluster info
  clustInfo = tmap$clustInfo
  clustMap = clustInfo$binMap
  
  ##create edge definition matrix
  ## - structured as 3 edges per cluster: 1 - left, 2 - bottom left, 3 - bottom right.
  ## -- we add one more row of padding for the top and rightmost edges.
  ## the two columns represent the two clusters that use that edge
  plotEdge = matrix(0, nrow=nrow*ncol*3, ncol=2)
  
  ##edge definitions (cell*3 + nVec = c(l, bl, br, r, tr, tl))
  eVec.1 = c(-2, -1, 0, 1, ((ncol+1)*3)-1, ((ncol)*3))
  eVec.2 = c(-2, -1, 0, 1, ((ncol)*3)-1, ((ncol-1)*3))
  
  ##neighbor definitions (used for empty cells that we want to combine with others)
  nVec.1 = c(-1, ncol, ncol+1, 1, -ncol+1,-ncol)
  nVec.2 = c(-1, ncol-1, ncol, 1, -ncol,-ncol-1)
  
  
  ##iterate over each cell
  for(i in (ncol+1):(nrow*(ncol-1))){
    nr = (i-1)%/%ncol
    nc = (i-1)%%ncol
    ci = match(i,cell)
    cl = ifelse(is.na(ci),NA, clustMap[ci])
    
    ##if it's a buffer, skip
    if(nr <= 0 || nr >= (nrow-1)){next;}
    
    ##if it's an empty cell (not defined in 'cell') combine with the most common 
    ## neighboring cluster
    if(is.na(ci)){
      nbrs = i + nVec.1
      if(nr%%2==0){nbrs = i+nVec.2}
      nbrs = match(nbrs, cell)
      ##if it's adjacent to at least one filled cell
      if(any(!is.na(nbrs))){
        nrbs = nbrs[!is.na(nbrs)]
        nbr.cl = clustMap[nbrs]
        if(any(!is.na(nbr.cl))){
          cl = as.integer(names(which.max(table(nbr.cl))))
        }
      }
    }
    
    ##otherwise, fill in edge info
    if(!is.na(cl)){
      n = i*3 + eVec.1
      if(nr%%2==0){n=i*3+eVec.2}
      
      plotEdge[n[1:3],1] = cl
      plotEdge[n[4:6],2] = cl
    } 
  }
  
  plotEdge
}

#' Method for determining cluster adjacency
#' @param clustMap       the cluster map
#' @param neighborMap    map of neighboring cluster
#' 
findAdjacentClusters = function(clustMap, neighborMap){
  nclust = max(clustMap, na.rm=T)
  
  lapply( 1:nclust, function(i){
    inI = which(clustMap==i)
    nbrs = unique(unlist(neighborMap[inI]))
    adj = sapply(1:nclust, function(j){
      inJ = which(clustMap==j)
      any(nbrs %in% inJ)
    })
    adj[i] = F
    which(adj)
  })
}



######## Helper functions ############

#' function for converting i/j vectors to indexes in a dist matrix
distdex<-function(i,j=NULL,n){ #given row, column, and n, return index
  drop=NULL
  if(all(j<0)){drop=j; j=NULL}
  if(is.null(j)){
    j=(1:n)[c(-i,drop)];
    ii=(1:n)[c(-i,drop)];
    j[i>j]=i;ii[i<j]=i; i = ii
  }
  n*(i-1) - i*(i-1)/2 + j-i
}

#' functions for converting from the index in a dissimilarity structure to row/column 
#' indexes
column_index <- function( I, M ){
  m = M-1; i = I - 1 ##zero based indexing
  col = (-2*m - 1 + sqrt( (4*m*(m+1) - 8*i - 7) )) / -2;
  if( col == floor(col) ) col = col- 1;
  floor(col) + 1;
}

row_index <- function( I, M ){
  i=I-1; col = column_index( I, M) - 1;
  i - (M-1) * col + col*(col+1) / 2 + 2;
}

rc_index <- function( I, M ){
  i=I-1; col = column_index( I, M) - 1;
  row = i - (M-1) * col + col*(col+1) / 2 + 1;
  c(row+1, col+1)
}


#' Weighted distance function
#' 
#' @name dist_wtd
#' 
#' @export
dist_wtd <- function(x, wt, method="euclidean", diag=FALSE, upper=FALSE, p=2){
  if (!is.na(pmatch(method, "euclidian"))) 
    method <- "euclidean"
  METHODS <- c("euclidean", "maximum", "manhattan", "canberra", 
               "binary", "minkowski")
  method <- pmatch(method, METHODS)
  if (is.na(method)) 
    stop("invalid distance method")
  if (method == -1) 
    stop("ambiguous distance method")
  x <- as.matrix(x)
  
  if(!is.matrix(wt)){
    if(is.data.frame(wt)){wt <- as.matrix(wt)}
    else if(is.vector(wt) && length(wt) == ncol(x)){
      wt = matrix(wt,nrow(x),ncol(x), byrow=T)
    }
    else{stop("invalid wt parameter")}
  }
  if(any(dim(wt) != dim(x))){stop("size of wt does not match x")}
  
  N <- nrow(x)
  attrs <- if (method == 6L) 
    list(Size = N, Labels = dimnames(x)[[1L]], Diag = diag, 
         Upper = upper, method = METHODS[method], p = p, call = match.call(), 
         class = "dist")
  else list(Size = N, Labels = dimnames(x)[[1L]], Diag = diag, 
            Upper = upper, method = METHODS[method], call = match.call(), 
            class = "dist")
  .Call("Cdist_wtd", x, wt, method, attrs, p)
}


#' Add buffer to hexbin edges
#' 
addEdgeBuffer = function(bin){
  ##pull out info
  cell <- bin@cell - 1
  n <- as.integer(length(cell))
  cnt <- bin@count
  xbins <- bin@xbins
  bdim <- bin@dimen
  row <- bdim[1]
  col <- bdim[2]
  
  ##add 4 total rows/columns
  ncol <- col + 4
  nrow <- row + 4
  nmax <- ncol * nrow
  
  nr <- cell %/% col + 2
  nc <- cell %% col + 2
  cell.new = nr * ncol + nc + 1
  
  
  bin@xbins <- xbins + 4
  bin@xbnds <- addBit(bin@xbnds, f = 2/xbins)
  bin@ybnds <- addBit(bin@ybnds, f = 2/xbins)
  bin@dimen <- c(nrow, ncol)
  bin@cell <- as.integer(cell.new)
  
  if(!is.null(bin@cID)){
    cID = cell.new[match(bin@cID-1, cell)]
    bin@cID = as.integer(cID)
  }

  bin
}
addBit <- function(bnds, f = 0.05) bnds + c(-f, f) * diff(bnds)




##--------------------------------------@
############# Old Functions #############
##--------------------------------------@

#' R implementation of hclust_adj
hclust_adj_r <- function(d, adj, weighted=T){
  n = attr(d, "Size")
  
  merge = matrix(0, n-1, 2)
  height = rep(0, n-1)
  index = -(1:n)
  counts = rep(1, n)
  np = n
  
  
  
  ##run hierarchical clustering
  for(i in 1:(n-1)){
    
    
    ##find what to combine
    if(weighted){
      ##calculate cluster size weights
      wts = c(1,0.8)[1+(counts==1)]
      wts = (wts %*% t(wts))
      wts = wts[lower.tri(wts)]
      
      j = which.min(d*adj*wts)}
    else{
      j = which.min(d*adj)
    }
    if(d[j]==Inf){ ##if no clusters are adjacent, just join closest
      j = which.min(d)
    }
    js = rc_index(j, np)
    j1 = js[1]; j2 = js[2]
    
    ##combine
    merge[i,] = c(index[j1],index[j2])
    height[i] = d[j]
    
    if(i==(n-1)){break}
    
    j1.i = distdex(j1,,np); j1.i = c(j1.i[seq(1,j1-1,length.out = j1-1)], NA, j1.i[seq(j1,np-1,length.out = np-j1)])
    j2.i = distdex(j2,,np); j2.i = c(j2.i[seq(1,j2-1,length.out = j2-1)], NA, j2.i[seq(j2,np-1,length.out = np-j2)])
    
    ##update distance matrix
    d.new =  pmax(d[j1.i], d[j2.i])
    d[j1.i[-j1]] = d.new[-j1]
    d = d[-j2.i[-j2]]
    
    ##update adjacency matrix
    a.new =  pmin(adj[j1.i], adj[j2.i])
    adj[j1.i[-j1]] = a.new[-j1]
    adj = adj[-j2.i[-j2]]
    
    ##update indexes
    index[j1] = i; index = index[-j2]
    np = np-1
    ##update counts 
    counts[j1] = counts[j1] + counts[j2]
    counts=counts[-j2]
    
  }
  list(merge=merge, height=height, labels=1:n)
}


#' Create plots for different numbers of clusters, highlighting which clusters were 
#' combined from the previous step
#' 
#' @export
animateClustering = function(hc, folder,tmap, from=1,to=50,addBorders=FALSE){
  ks = from:to
  grps = cutree(hc, ks)
  clustLoc = tmap$binInfo$means
  
  ##digits for file names
  digits = floor(log10(length(ks)))+1
  ##largest number of clusters
  max.nc = max(apply(grps,2,function(x){length(unique(x))}))
  
  for(i in 1:length(ks)){
    
    ##open file
    frame = paste0(c(rep(0,digits-(floor(log10(i))+1)), i),collapse = "")
    fn = file.path(folder,paste0(frame, ".png"))
    
    
    if(i==1){col=sample(rainbow(max(grps[,i])))[grps[,i]]}
    else{
      if(from<to){
        split = which(colSums(table(grps[,i],grps[,i-1])>0)>1)
        col = rep("#999999",nrow(clustLoc))
        col[grps[,i-1]==split] = "#FF0000"
      }else{
        split = which(colSums(table(grps[,i-1],grps[,i])>0)>1)
        col = rep("#999999",nrow(clustLoc))
        col[grps[,i]==split] = "#FF0000"
      }
    }
    
    png(fn, 500,500)
    plot(clustLoc,col=col,pch=16)
    if(addBorders){
      tmap$clustInfo$binMap = grps[,i]
      plotClusterEdges(tmap)
    }
    dev.off()
  }
}

#' Adjust the clusters by hand
#' 
# handCluster <- function(tmap, pch = 19, pallette = NULL, ...)
# {
#   if(is.null(tmap$binInfo) || is.null(tmap$cluster)){
#     stop("cluster information not found")
#   }
#   tools = list()
#   tools$selectClust <- list("selectClust", "gWidgetsRGtk2-symbol_square",
#       label="Add/Drop Clusters", accelerator="F5",
#       tooltip="add or remove clusters to selection",
#       callback=function(widget, playState){
#         repeat{
#           foo <- playSelectData(playState,
#                                 prompt=paste(
#                                   "Click to select/add a Cluster.",
#                                   "ctrl-click to remove.",
#                                   "Right-click to stop."))
#           if(is.null(foo)) return()
#           pts = foo$which
#           if (foo$modifiers & RGtk2::GdkModifierType["control-mask"]) {
#             ##control-click: drop cluster
#             cl = playState$env$dat$clustMap[pts]
#             pts = which(playState$env$dat$clustMap %in% cl)
#             prev = playGetIDs(playState)
#             pts = prev[!(prev %in% pts)]              
#             playSetIDs(playState, pts, redraw = T)
#             
#           } else{
#             ## click: add cluster
#             cl = playState$env$dat$clustMap[pts]
#             pts = which(playState$env$dat$clustMap %in% cl)
#             prev = playGetIDs(playState)
#             pts = unique(c(prev,pts))              
#             playSetIDs(playState, pts, redraw = T)
#           }
#         }
#       }
#   )
#   tools$selectPts <- list(name="Pts", stock_id="gWidgetsRGtk2-symbol_dot",
#       label="Add/Drop Points", accelerator="F6",
#       tooltip="add or remove points to selection",
#       callback=function(widget, playState){
#         repeat{
#           foo <- playSelectData(playState,
#                                 prompt=paste(
#                                   "Click to add points",
#                                   "ctrl-click to remove.",
#                                   "hold to select rectangle",
#                                   "Right-click to stop."))
#           if(is.null(foo)) return()
#           pts = foo$which
#           if (foo$modifiers & RGtk2::GdkModifierType["control-mask"]) {
#             ##control-click: drop points
#             prev = playGetIDs(playState)
#             pts = prev[!(prev %in% pts)]              
#             playSetIDs(playState, pts, redraw = T)            
#           } else{
#             ## click: add cluster
#             prev = playGetIDs(playState)
#             pts = unique(c(prev,pts))              
#             playSetIDs(playState, pts, redraw = T)
#           }
#         }
#       }
#   )
#   tools$refresh <- list("refresh", label="Refresh Colors", accelerator="F7",
#       tooltip="Generate new color scheme for clusters",
#       callback=function(widget,playState){
#         with(playState$env,{
#           pallette = sample(rainbow(nclust+10))
#           cols = pallette[dat$clustMap]
#         })
#         playReplot(playState)
#       }
#   )
#   tools$newClust <- list("newClust", label="New Cluster", accelerator="F8",
#       tooltip="create new cluster from selection",
#       callback=function(widget, playState){
#         pts = playGetIDs(playState)
#         if(is.null(pts)){return()}
#         playState$env$pts = pts
#         
#         with(playState$env, {
#           srcClust = unique(dat$clustMap[pts])
#           ##find which clusters are completely selected
#           hasAll = rep(F, length(srcClust))
#           for(i in 1:length(srcClust)){
#             if(  sum(dat$clustMap[pts]== srcClust[i]) == 
#                  sum(dat$clustMap     == srcClust[i])){
#               hasAll[i] = TRUE
#             }
#           }
#           ##find index for new cluster
#           newClust = nclust + 1
#           if(any(hasAll)){newClust = srcClust[hasAll][1]}          
#           ##update nclust
#           nclust = nclust - sum(hasAll) + 1
#           ##update cluster info
#           dat$clustMap[pts] = newClust
#           ##if we run off the edge of the pallette, regenerate
#           if(nclust>length(pallette)){pallette = sample(rainbow(nclust+10))}
#           ##update colors
#           cols = pallette[dat$clustMap]
#           pts = NULL
#         })
#         playClear(playState, "brushed")
#         playReplot(playState)
#       }
#   )
# 
#   try(playwith({
#     nclust = tmap$binInfo$nclust
#     dat = as.data.frame(tmap$binInfo$means); colnames(dat) = c("x","y")
#     dat$clustMap = tmap$binInfo$clustMap
#         
#     if(is.null(pallette)){pallette = sample(rainbow(nclust+10))}
#     cols = pallette[dat$clustMap]
#     xyplot(y~x, dat, pch=pch, col=cols)
#   }, tools = tools), silent=T)#, update.actions=list(selector))
#     
# #     sel <- rep(FALSE, length(x)); res <- integer(0)
# #     while(sum(sel) < n) {
# #       ans <- identify(x, y, n = 1, plot = FALSE, ...)
# #       if(!length(ans)) break
# #       
# #       if(method=='combine'){
# #         ans = which(clustMap == clustMap[ans])
# #       }
# #       points(x[ans], y[ans], pch = pch, col=ifelse(sel[ans], cols[ans],"#C0C0C0"))
# #       if(all(!sel[ans])){points(x[ans], y[ans], pch = pch, col="black",cex=0.9)}
# #       sel[ans] <- !sel[ans]
# #     }
# #     if(!any(sel)) break
# #     if(method=='split'){
# #       newClust.map[sel] = max(newClust.map)+1
# #     }else if(method=='combine'){
# #       newClust.map[sel] = min(newClust.map[sel])
# #     }
# #     
# #   }
# #   newClust.map
# }


