library("hexbin", "gplots")

rainbowFn = colorRamp(
  c(colorRampPalette(c("#000088","#0000FF"))(8),
    colorRampPalette(c("#0011FF","#00FFFF"))(6),
    colorRampPalette(c("#00FFEE","#00FF00"))(3),
    rev(rainbow(20))[-(1:13)]
  ))
smooth.wts = c(30,7,3)

densityCols = function(n, pctile.trim=100, norm=F, pctMax=NULL){
  col.fn = rainbowFn
  
  maxCol=quantile(n, probs = pctile.trim/100, na.rm=T)
  if(norm){maxCol=pctMax}
  
  col = (n - min(n,na.rm = T))/(maxCol - min(n, na.rm=T))
  col[col>1] = 1
  col.mat = col.fn(col[!is.na(col)])
  
  col.avg = rep("#FFFFFF", length(col))
  col.avg[!is.na(col)] = rgb(col.mat, maxColorValue = 255)
  col.avg
}

#' creates a new plot that adds a color key based on the provided range
#' 
#' @export
colorKey = function(range, at=NULL, labels=NULL, colFn=colorRampPalette(c("green","black","red"))){
  ncol<-25
  keyCols <- colFn(ncol)
  x<-seq(0.1,0.9,length.out=ncol+1)
  plot.new();
  for(i in 1:ncol){rect(x[i],0.4,x[i+1],0.6,col=keyCols[i],border=NA)} 

  if(is.null(at) & !is.null(labels)){stop("at must be specified if labels is provided")}
  if(is.null(at)){at = pretty(seq(range[1],range[2], length.out=5))}
  if(is.null(labels)){labels = at}
  at = (at - range[1])/(range[2]-range[1]) * 0.8 + 0.1

  labels = labels[at>=0.1 & at<=0.9]; at = at[at>=0.1 & at<=0.9]

  for(i in 1:length(at)){
    lines(c(at[i],at[i]),c(0.4,0.35))
  }
  text(at,rep(0.3,length(at)), label=labels)

}


#' Custom color Hexplot
#' @author Chris Bolen
#' @name hexplot
#' @description high density plotting using hexagonal binning with overlaid color
#'              information.
#' 
#' @param x,y                the coordinates of points in the plot, or a hexbin object. 
#' @param hex                a hexbin object
#' @param col                vector of same length as X describing coloring information
#'                           for each point. If values are between 0 and 1, the values are
#'                           averaged across all points in the bin, and bin colors are 
#'                           calculated using col.fn. Otherwise, values must be valid 
#'                           color definitions. Specifying NULL will result in rainbow 
#'                           colors being used, as in the main hexplot function. 
#' @param col.fn             a colorRamp function which takes values from 0 to 1 and 
#'                           returns RGB specifications. Default is a colorRamp from Green -> Black -> Red.
#' @param count.transform    function; the transformation to apply to counts before 
#'                           calculating colors. Useful for visualizing lower tail of 
#'                           distribution.
#' @param xbins              the number of bins partitioning the range of x.
#' @param pctile.trim        value defining the maximum percentile for calculating colors.
#'                           Only used if col=NULL (see details).
#' @param min.grp            Minimum number of points to combine together into a bin
#' @param cex                cex of largest map points. (see details for information on 
#'                           how the cex for each point is calculated). 
#' @param calc.cex           boolean; if true, then cex sizes will be calculated based on
#'                           number of points in each bin. Used for displaying density if
#'                           custom colors are provided
#' @param calc.alpha         boolean; if true, then alpha channels will be added to show
#'                           density. Use as an alternative to calc.cex.
#' @param alpha.ranges       numeric vector of length 2; describes the alpha value for the
#'                           minimum and median bin size (see details).
#' @param smooth             boolean; if true, smooth.hexbin is called to smooth densities
#'                           of adjacent hexbins
#' @param norm               boolean; if true, convert counts to percentage before 
#'                           normalization and calculating colors.
#' @param pctMax             Alternative to pctile.trim. Specifies the maximum seperable 
#'                           value as a percent of the total. Useful for consistency of
#'                           color across multiple plots. Only used if norm=T
#' @param plot               logical; if false, plotting will be supressed, and the hexbin
#'                           object will be returned instead.
#' @param add                logical; if false, plot.new will not be called
#' @param xlim,ylim          x and y limits of the plotting region.
#' @param xlab,ylab          x and y axis labes
#' @param log                character string specifying which axes should be log-scale.
#' @param pch,...           default plotting parameters to pass to 'plot'
#' 
#' @details
#' This function is built off "hexbin" package as an alterative to their less versatile
#' plotting function. The plot is created by breaking the plotting region up into equally
#' sized hexagons, counting the number of points inside each, and then using the count to 
#' determine the color of that bin. 
#' 
#' In order to make the hexbin package interface more cleanly with the default plot
#' function, the hexagons are represented by single points at the midpoint of each 
#' hexagonal region rather than as hexagons, like those used in the actual hexbin package.
#' 
#' In some cases, you will have large outlier bins which dominate the color range.
#' In such cases it is useful to set the max of the color range to a specific percentile
#' by setting the pctile.trim parameter. This will set all bins in that percentile and
#' above to the same color, enhancing the separation between bins below that percentile.
#' 
#' If necessary, additional color information can be supplied for each individual 
#' point, supplanting the default point density colors. This can be done by supplying a 
#' vector of colors to the \code{col} parameter. 
#' The \code{col} parameter takes a vector of colors for each individual point, and the 
#' final plotted color is defined by averaging the RGB values of all points in the bin. 
#' Alternatively, \code{col} can be a numeric vector with values ranging from 0 to 1, 
#' representing some continuous property of the map. In this case, the numeric values are
#' averaged for each point in the bin and are then converted to colors using the colorRamp
#' function defined by \code{col.fn} The result is a heatmap-like visualization of the 
#' provided parameter.
#' 
#' In cases where \code{col}, instead of displaying density information using color, hexbin 
#' densities are displayed using either varying point sizes (if calc.cex is TRUE), or 
#' using varying levels of transparency (if calc.alpha is TRUE). 
#' Point sizes and alpha parameters for the plot are calculated in a similar way as the 
#' colors in hexplot. Bin sizes are first calculated and transformed (using the 
#' \code{count.transform} function). If \code{calc.cex} is true,
#' the bin sizes are then scaled such that the average bin has a value of \code{cex/2}. 
#' Then, all bins greater than \code{cex} are set to \code{cex}.
#' If \code{calc.alpha} is TRUE, the counts will be scaled such that the minimum value is
#' equal to \code{alpha.ranges[1]}, and the mean value will be equal to 
#' \code{alpha.ranges[2]}. Then, all alpha values greater than 1 are set to 1.
#'  
#' @return hexplot.default invisibly returns the generated hexbin object
#' 
#' @examples 
#' ##create dataset
#' data = rbind(matrix(rnorm(40000), ncol=2),
#'              matrix(rnorm(40000), ncol=2) + 2,
#'              matrix(c(rnorm(20000)+2,rnorm(20000)-2), ncol=2)
#'             )
#'             
#' ##standard density-colored hexplot
#' hexplot(data, xbins=100)
#' 
#' ##altering the color parameters
#' hexplot(data, xbins = 100, count.transform=NULL, pctile.trim=95)
#' 
#' ##add custom color information
#' cols = rep(c("red","green","blue"), each=20000)
#' hexplot(data, xbins=100, col=cols)
#' 
#' @export
hexplot = function(x,...) UseMethod("hexplot")
#' @rdname hexplot
#' @export
hexplot.default = function(x, y=NULL, xbins=150, smooth=F, plot=T, min.grp = 5,
                           xlim=NULL, ylim=NULL, xlab=NULL, ylab=NULL, log="", ...){
  
  ##check parameters
  if(smooth && min.grp>1){
    warning("Cannot split groups when smooth=TRUE. Changing min.grp to 1");
    min.grp=1
  }
  
  ##make sure we have an x and y
  if(is.null(y)){
    ##if x is a matrix
    if((class(x)=="matrix" || class(x)=="data.frame") && ncol(x)>1){
      if(is.null(xlab)){
        if(!is.null(colnames(x))){xlab=colnames(x)[1]}
        else{xlab = paste0(substitute(x), "[,1]")}
      }
      if(is.null(ylab)){
        if(!is.null(colnames(x))){ylab=colnames(x)[2]}
        else{ylab = paste0(substitute(x), "[,2]")}
      }
      
      y = x[,2]; x = x[,1]
      
      ##else, assume that x is a vector from 1:N
    }else{
      if(is.null(xlab)){xlab="count"}
      if(is.null(ylab)){ylab=substitute(x)}
      
      y = x
      x = 1:length(y)
    }
  }else{
    if(is.null(xlab)){xlab=substitute(x)}
    if(is.null(ylab)){ylab=substitute(y)}
  }
  if(is.null(xlim)){xlim=range(x,na.rm=T)}
  if(is.null(ylim)){ylim=range(y,na.rm=T)}
  
  ##remove points outside xlim & ylim
  x = x[y>=ylim[1] & y<=ylim[2]]; y = y[y>=ylim[1] & y<=ylim[2]]
  y = y[x>=xlim[1] & x<=xlim[2]]; x = x[x>=xlim[1] & x<=xlim[2]]
  
  ##log-transform data
  if(grepl("x",log)){
    x.l = log10(x); 
    x.l[x<=0] = NA; 
    if(xlim[1]<=0){xlim[1] = min(x[x>0])}
    x = x.l
  }
  if(grepl("y",log)){
    y.l = log10(y); 
    y.l[y<=0] = NA; 
    if(ylim[1]<=0){ylim[1] = min(y[y>0])}
    y = y.l
  }
  
  ##get bins
  hex = hexbin(x,y,xbins=xbins,IDs=T,xbnds=xlim,ybnds=ylim)
  if(smooth){
    sm.hex = smooth.hexbin(hex,smooth.wts)
    sm.map = smoothbin.map(hex)
    sm.hex@cID = as.integer(sm.map$new[match(sm.hex@cID, sm.map$old)])
    hex = sm.hex
  }
  if(plot){
    hexplot.hexbin(hex,x,y, min.grp=min.grp, xlim=xlim, ylim=ylim,xlab=xlab, ylab=ylab,
                   smooth=F, log=log,...)
  }
  invisible(hex)
}
#' @rdname hexplot
#' @export
hexplot.tmap = function(x, xlab=NULL, ylab=NULL, ...){
  if(is.null(x$tSNE.map)){stop("tSNE.map is not defined")}
  if(is.null(xlab)){xlab="tSNE dim1"}
  if(is.null(ylab)){ylab="tSNE dim2"}
  
  y = x$tSNE.map[,2];x = x$tSNE.map[,1];
  
  hexplot.default(x,y, xlab=xlab, ylab=ylab,...) 
}
#' @rdname hexplot
#' @export
hexplot.hexbin = function(hex, x=NULL,y=NULL, col=NULL, binCol=NULL, col.fn = NULL,
                          count.transform=log2, pctile.trim=NULL, min.grp=ifelse(is.null(x), 1,5), 
                          cex=1, calc.cex=F, calc.alpha=!is.null(col), alpha.ranges=c(0.2,0.6),
                          smooth=F, norm=F, pctMax=NULL, 
                          log="", xlab=NULL, ylab=NULL,add=F, pch=16, ...){
  
  ##if we're ungrouping, make sure x and y information exists
  if(is.null(x) && min.grp>1){stop("original points (x and y) must be provided if min.grp>1")}
  
  ##if we need an x and y, make sure we have both
  if(min.grp>1 && is.null(y)){
    ##if x is a matrix
    if((class(x)=="matrix" || class(x)=="data.frame") && ncol(x)>1){
      if(is.null(xlab)){xlab=colnames(x)[1]}
      if(is.null(ylab)){ylab=colnames(x)[2]}
      
      y = x[,2]; x = x[,1]
      
      ##else, assume that x is a vector from 1:N
    }else{
      if(is.null(xlab)){xlab="count"}
      if(is.null(ylab)){ylab=substitute(x)}
      
      y = x
      x = 1:length(y)
    }
  }
  
  ##define x and y labels (if not already defined)
  if(is.null(xlab)){xlab=hex@xlab}
  if(is.null(ylab)){ylab=hex@ylab}
  
  ##define col.fn
  if(is.null(col.fn)){
    col.fn = colorRamp(c("green", "black", "red"))
  }
  
  ##check that parameters exist
  if(norm && is.null(pctMax)){stop("pctMax must be provided if norm is TRUE")}
  if(is.null(pctile.trim)){pctile.trim = 100}
  
  ##initialize col.avg
  col.avg = NULL
  
  ##smooth the hexbin object
  if(smooth){
    sm.hex = smooth.hexbin(hex,smooth.wts)
    sm.map = smoothbin.map(hex)
    sm.hex@cID = as.integer(sm.map$new[match(sm.hex@cID, sm.map$old)])
    hex = sm.hex
  }
  
  ##get hex locs
  xy = hcell2xy(hex)
  x.c = xy$x; y.c = xy$y
  ##and un-log transform it (if it was transformed earlier)
  if(grepl("x",log)){x.c = 10^(x.c)};if(grepl("y",log)){y.c = 10^(y.c)}
  n = hex@count
  n[n==0] = NA
  cID = hex@cID
  
  ##ungroup all the bins with less than min.grp number of points
  if(!is.na(min.grp) & min.grp>1){
    isSmall = which(n<=min.grp)
    toAdd = which(cID %in% hex@cell[isSmall])
    ##remove the small bins
    x.c = x.c[-isSmall]; y.c = y.c[-isSmall]; n = n[-isSmall]; 
    ##correct the remaining counts
    n = n - min.grp
    ##and add the original points back
    x.c = c(x.c, x[toAdd]); y.c = c(y.c, y[toAdd])
    n = c(n, rep(1,length(toAdd)))
    cID[toAdd] = (max(cID)+1):(max(cID)+length(toAdd))
  }
  
  
  ##normalize bin sizes
  if(norm){n = n/sum(n, na.rm=T)}
  
  ##if add is false, call plot.new
  if(!add){
    plot(x.c, y.c,type="n",xlab=xlab,ylab=ylab,log=log,...)
  }
  
  ##figure out how big our point size should be
  minPtRadius.in = 0.004 ##smallest possible point size (in inches)
  ##get plotting region size
  dsize = plotSize()
  ##divide by number of bins
  binRadius.in = max(dsize/hex@dimen)/2 * 1.15 * cex
  if(binRadius.in<=minPtRadius.in){
    binRadius.in=minPtRadius.in;
    calc.cex=F
  }
  cex.min = minPtRadius.in/binRadius.in
  cex.mid = cex.min+(cex-cex.min)/2
#   if(cex.min*3.1>1){
#     ##cat("cex too small. Increasing cex\n"); 
#     binRadius.in = minPtRadius.in*3.1
#     cex = binRadius.in / (max(dsize/hex@dimen)/2)
#   }
  
  
  
  ##if we need to, calculate the bin density
  if((is.null(col) && is.null(binCol)) || calc.cex || calc.alpha){
  
    # fetch the hex count
    if(!is.null(count.transform)){
      n  = count.transform(n)
      pctMax = if(!is.null(pctMax)) count.transform(pctMax)
    }

    # if all the bin sizes are equal (i.e. all 1, just make everything the same size and color)
    if(min(n,na.rm=T)==max(n,na.rm=T)){
      cexes = rep(cex, length(n))
      if(is.null(binCol) && is.null(col))
        col.avg = rep(rgb(rainbowFn(0), maxColorValue = 255),length(n))
    }else{
      ##scale n between cex.min and cex.mid
      if(calc.cex){
        cexes  = n  - min(n,na.rm=T)
        cexes  = cexes  / mean(cexes,na.rm=T) * cex.mid + cex.min
        cexes[cexes>1] = 1
      }
      
      ##calc alphas by scaling n between 0.2 (min alpha) and 0.6 (median alpha)
      if(calc.alpha){
        alphas  = n  - min(n,na.rm=T)
        alphas  = alphas  / mean(alphas,na.rm=T) * alpha.ranges[2] + alpha.ranges[1]
        alphas[alphas>1] = 1
      }
      
      # if col is null, color by density
      if(is.null(col) && is.null(binCol)){
        col.avg = densityCols(n, pctile.trim=pctile.trim, norm=norm, pctMax=pctMax)
      }

    }
  }
  
  ##if we're not calculating cex, set them all to "cex"
  if(!calc.cex){cexes = rep(1, length(n))}
  
  ##if binCol is defined, use that for col.avg
  if(!is.null(binCol)){col.avg=binCol}
  
  ##if cols are provided, average colors across each bin
  if(is.null(col.avg)){
    if(length(col) != length(cID)){stop("length of col must match length of x and y")}
    
    # if cols is a vector from 0 to 1, average them together
    if(is.numeric(col) && min(col, na.rm=T)>=0 && max(col, na.rm=T)<=1){
      col.avg = sapply(split(col, cID),function(z){
        rgb(col.fn(mean(z)), maxColorValue = 255)
      })
    }else{ ##else, if vectors are colors of some kind
      col.avg = sapply(split(col, cID),function(z){
        if(all(is.na(z))){return(NA)}
        rgbMat = col2rgb(z[!is.na(z)], alpha=T)
        alpha = rgbMat["alpha",]/255
        rgbMat = t(rgbMat[1:3,])*alpha
        rgb(t(colSums(rgbMat)/sum(alpha)), maxColorValue = 255)
      })
    }
  }
  ##add the alpha channel to the colors (if we specified it)
  if(calc.alpha){col.avg = paste0(col.avg, as.hexmode(as.integer(alphas*255)))}
  
  ##plot points
  symbols(x.c,y.c, circles=cexes, inches=binRadius.in, add=T, fg="#00000000", bg=col.avg)
  
}



#' Downsample a 2-d map 
#' @export
hex.downSample = function(x,y=NULL, sampleFn=function(x){ceiling(sqrt(x))}, xbins=100){
  if(is.null(y)){
    if((class(x)=="matrix" || class(x)=="data.frame") && ncol(x)>1){
      y = x[,2]; x = x[,1]
    }else{
      stop("No Y variable found for down-sampling")
    }
  }
  
  hex = hexbin(x,y,xbins=xbins, IDs=T)
    
  toSelect = sampleFn(hex@count)
  selected = unlist(lapply(1:length(hex@cell), function(i){
    pts = which(hex@cID==hex@cell[i])
    resample(pts, size=toSelect[i], replace=F)
  }))
  return(selected)
  
}
resample <- function(x, ...) x[sample.int(length(x), ...)]


######### Helper Functions #######
plotSize <- function(){
  dev = dev.size("in")
  mfrow = par("mfrow")
  dev = dev/rev(mfrow)
  dev[1] = dev[1] - sum(par("mai")[c(2,4)])
  dev[2] = dev[2] - sum(par("mai")[c(1,3)])
  return(dev)
}
##map smoothbin cells back to regualr cells
smoothbin.map <- function(bin){
  if (!is(bin, "hexbin")) 
    stop("first argument must be a hexbin object")
  cell <- bin@cell - 1
  n <- as.integer(length(cell))
  cnt <- bin@count
  xbins <- bin@xbins
  bdim <- bin@dimen
  row <- bdim[1]
  col <- bdim[2]
  ncol <- col + 4
  nrow <- row + 4
  nmax <- ncol * nrow
  sm <- rep.int(0, nmax)
  nr <- cell%/%col + 2
  nc <- cell%%col + 3
  pad <- rep.int(0:0, nmax - n)
  cell <- c(nr * ncol + nc, pad)
  cnt <- c(cnt, pad)
  
  old = c(bin@cell,rep(NA,nmax-n))
  return(list(old=old,new=cell))
}