#' Plot matrix of markers
#' 
#' @name plotMarkers
#' @description Generate hexplot maps for expression of a set of markers
#' 
#' @param tmap          the tmap object to be plotted
#' @param markerList    list of markers to be included in the plot. If null, all markers 
#'                      will be plotted.
#' @param col.lim       vector of length 2; describes the minimum and maximum values used 
#'                      when calculating the color range. Values outside this range will
#'                      be set to the closest color inside the range, as appropirate. If 
#'                      null, col.lim is set to \code{range(tmap$markerVals)}.
#' @param mfrow         vector of length 2, to be passed on to par. Determines number of 
#'                      rows/columns to include in each plot. If not supplied, an 
#'                      appropriate mfrow will be determined automatically.
#' @param xbins         number of bins that divide the x axis. See hexplot for details
#' @param min.grp       minimum bin size for grouping points. See hexplot for details 
#' @param rasterize     boolean; should the image be rasterized (converted to png before 
#'                      plotting)? Useful to reduce file size when plotting to pdf 
#'                      (or other vector-based device).
#' @param col.fn        a colorRamp function which takes values from 0 to 1 and returns 
#'                      RGB specifications. Default is a colorRamp from Green -> Black -> Red.
#' @param colKey        named vector indicating the values for the color key
#' @param ...           additional parameters to be passed to hexplot
#' 
#' @details 
#' The \code{plotMarkers} function creates a matrix of hexplots, colored by the expression
#' of the markers in \code{tmap$markerVals}. The colors are determined using a 
#' \code{\link[RColorBrewer]{colorRamp}} function from the \code{RColorBrewer} package, 
#' which can be specified via the '\code{col.fn}' parameter (see 
#' \code{\link[RColorBrewer]{colorRamp}} for details on supplying alternative color fns). 
#' By default, colors will range from green (lowest value) to red (highest value), with 
#' the lowest and highest values defined by \code{col.lim}. Any values outside the range of 
#' \code{col.lim} will be displayed as either completely red or completely green.
#' 
#' If \code{rasterize} is \code{TRUE}, each individual plot will be converted to a raster
#' image and plotted using the \code{rasterImage} function. This is especially important
#' if the plots are being saved to a vector-based format, such as a pdf, because the large
#' number of individual points in each graph will make the pdf file impractically large.
#' 
#' @export 
plotMarkers <- function(tmap, markerList=NULL, col.lim=NULL,mfrow=NULL, 
                        addEdges=!is.null(tmap$binInfo),xbins=100, min.grp=5,
                        rasterize = (names(dev.cur())=="pdf"), 
                        col.fn = NULL, colKey = NULL,...){
  ##first, check inputs
  if(class(tmap)!="tmap" || is.null(tmap$tSNE.map)){
    stop("tmap must be a valid 'tmap' object")
  }
  if(is.null(markerList)){
    markerList = colnames(tmap$markerVals)
  }
  if(is.numeric(markerList)){
    markerList = colnames(tmap$markerVals)[markerList]
  }
  if(!all(markerList %in% colnames(tmap$markerVals))){
    missing = markerList[!(markerList %in% colnames(tmap$markerVals))]
    stop("Markers in markerList were not found in tmap: ",paste(missing,collapse=", "))
  }
  if(is.null(col.lim)){ 
    col.lim = range(tmap$markerVals, na.rm=T)
  }
  if(is.null(mfrow)){ ##automatically determine mfrow if not supplied
    mfrow=c(1,1)
    l = length(markerList)
    if(l>3){mfrow[1] = 2}
    if(l>1){mfrow[2] = ifelse(l %in% c(2,4), 2, 3)}
  }
  plotPerPage = mfrow[1]*mfrow[2]
  
  if(!is.null(colKey)){
    colorKey(col.lim, at=colKey, labels=names(colKey))
  }
  
  
  ##open new device for rasterizing
  if(rasterize) {
    dev=dev.cur(); ds = dev.size("in"); 
    op= par(mai=c(0,0,0,0), mfrow=c(1,1))
    fn = tempfile()
    png(fn,ds[1],ds[2], units="in", res = 200)
    par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
  }else{
    ##set plotting parameters
    op = par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
  }
  
  ##plot all markers
  for(i in 1:length(markerList)){
    n = markerList[i]
    
    ##calculate colors
    cols = tmap$markerVals[,n]; 
    cols=cols-col.lim[1]; 
    cols=cols/col.lim[2]; 
    cols[cols>1]=1
    cols[cols<0]=0
    ##call hexplot
    hexplot(tmap,col=cols, xbins=xbins,axes=F, main=n, min.grp=min.grp, col.fn=col.fn,...)
    box()
    
    if(addEdges){tmap = plotClusterEdges(tmap, line.col="blue")}
    
    if(rasterize && i%%plotPerPage==0){
      dev.off()
      plotPNG = readPNG(fn)
      dev.set(dev)
      plot(c(0,1),c(0,1),type="n");rasterImage(plotPNG,0,0,1,1)
      png(fn,ds[1],ds[2], units="in", res = 200)
      par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
    }
  }
  ## if rasterize, plot last page
  if(rasterize){
    dev.off();
    dev.set(dev)
    if(i%%plotPerPage!=0){
      plotPNG = readPNG(fn)
      plot(c(0,1),c(0,1),type="n");rasterImage(plotPNG,0,0,1,1)
    }
  }
  
  ##reset plotting parameters back to original
  par(op)
}

#' Plot cluster information
#' @name plotClusters
#' @description generate cluster maps using the hexplot function
#' 
#' @param tmap           the tmap object to plot. Must contain cluster info
#' @param labelCluster   boolean; if true, cluster numbers will be annotated
#' @param labelCol       the color for the cluster labels; default is to use the same
#'                       color as the clusters.
#' @param addEdges       either a boolean or a numeric vector; if the numeric vector is 
#'                       defined, the edges for those clusters will be plotted. 
#'                       Alternatively, specifying TRUE will specify that all edges
#'                       should be plotted.
#' @param cluster.col    vector of colors for each cluster. Length must be equal to the 
#'                       number of clusters
#' @param ...            additional parameters to be passed to hexplot
#' 
#' @details 
#' By default, the colors for the clusters will be determined by the 
#' \code{\link[grDevices]{rainbow}} function. The function will generate a number of 
#' colors equal to the number of clusters, and then the rainbow colors will be randomly
#' rearranged to ensure that nearly identical colors are not right next to each other. 
#' One consequence of this is that each call to \code{plotClusters} will generate a 
#' different color scheme. Therefore, the only way to ensure consistency of colors between
#' multiple plots is to supply a vector of colors to \code{cluster.col}. In order to make
#' this easier using the function, the randomly generated color vector used in the plot
#' is invisibly returned, meaning that you can store the color vector from one call and
#' use it in any subsequent calls to the function.
#' 
#' @return invisibly returns the color scheme for the clusters.
#' @export
plotClusters <- function(tmap, labelCluster=TRUE, labelCol=NULL, 
                         addEdges=!is.null(tmap$binInfo), 
                         cluster.col=NULL, ...){
  ##check that clustering has been run
  if(is.null(tmap$clustInfo)){stop("Clustering information not found")}
  ##pull out cluster info
  clustInfo = tmap$clustInfo
  
  nclust = clustInfo$nclust
  
  ##calculate cluster means 
  comb.center = sapply(1:nclust, function(i){
    apply(tmap$tSNE.map[which(clustInfo$cluster==i),,drop=F],2,mean)
  })
  comb.center = t(comb.center)
  
  ##generate colors
  if(is.null(cluster.col))  cluster.col = sample(rainbow(nclust))
  cols = cluster.col[clustInfo$cluster]
  cols[is.na(cols)] = "#888888"
  hexplot(tmap, col = cols,...)
  ##plot cluster edges
  if(!is.null(addEdges) && (is.numeric(addEdges) || (is.logical(addEdges) && addEdges) )){
    if(is.numeric(addEdges)){plotClusterEdges(tmap,clusters = addEdges)}
    else{plotClusterEdges(tmap)}
  }
  ##add cluster labels
  if(labelCluster){
    points(comb.center,pch=16,cex=2.7,col="black")
    points(comb.center,pch=16,cex=2.5,col="white")
    if(is.null(labelCol)){labelCol = cols[match(1:nrow(comb.center), clustInfo$cluster)]}
    
    text(comb.center, labels = 1:nrow(comb.center), col=labelCol)
  }
  invisible(cluster.col)
}

#' Add cluster edges to map
#' @param tmap      A tmap object containing clustering information
#' @param clusters  numeric vector; a list of which clusters for which borders should be
#'                  plotted.
#' @param add       boolean; if TRUE, cluster outlines will be added to the current plot.
#' @param lwd       numeric; width of the line separating colors
#' @param line.col  the color of cluster edge lines. Default is black
#' @param ...       additional parameters to be passed to \code{hexplot} if add=T
#' 
#' @export
plotClusterEdges <- function(tmap, clusters=NULL, add=T, lwd=2, line.col=NULL, ...){
  ##standard checks
  if(is.null(tmap$clustInfo)){stop("Clustering information not found")}
  ##pull out cluster info
  clustInfo = tmap$clustInfo
  
  ##edit input
  if(is.null(line.col)){line.col=1}
  
  ##if add=FALSE, make a standard hexplot first
  if(!add){hexplot(tmap,...)}
  
  ##check clusters
  if(is.null(clusters)){clusters = 1:clustInfo$nclust}
  
  ##find the edges of the clusters
  if(is.null(clustInfo$clustEdges)){
    edgeMat = findClustEdges(tmap)
    tmap$clustInfo$clustEdges = edgeMat
  }else{edgeMat = clustInfo$clustEdges}
  
  edgesToPlot = edgeMat[,1] != edgeMat[,2] & 
               (edgeMat[,1] %in% clusters | edgeMat[,2] %in% clusters) 
  
  ####
  ##figure out edge locations
  
  hexes = tmap$binInfo$hexbin
  nrow = hexes@dimen[1]
  ncol = hexes@dimen[2]
  ##bin width/height
  dxy = getBinSize(hexes)
  ##bin edges (add to bin centers to get pts)  
  edges = getBinEdges(dxy)
  ##bin centers  
  xy.c = getBinCenters(hexes)
  x.c = xy.c[,1]; y.c = xy.c[,2]
  
  ####
  ##iterate over each point and plot edges
  for(i in 1:(nrow*ncol)){
    
    #figure out which edges to plot
    toPlot = edgesToPlot[(i*3-2):(i*3)]
    if(any(toPlot)){
      for(e in edges[toPlot]){
        x = x.c[i] + e$x
        y = y.c[i] + e$y
        lines(x,y,lwd=lwd,col=line.col,xpd=T)
      }
    }
  }
  invisible(tmap)
}

#' Heatmap of marker expression in each cluster
#' 
#' @name markerHeatmap
#' @description Generates a summary of average marker expression for each cluster and
#'             displays as a heatmap
#'            
#'            
#' @export
markerHeatmap <- function(tmap,trace='n', key.xlab="Average Expression",...){
  ##pull out stuff from tmap
  if(is.null(tmap$clustInfo)){stop("cluster information not found")}
  clustInfo = tmap$clustInfo
  markerVals = tmap$markerVals
  nclust = clustInfo$nclust
  
  ##calculate average marker expression
  avgExpression = sapply(1:nclust, function(i){
    inI = which(clustInfo$cluster == i)
    colSums(markerVals[inI,])/length(inI)
  })
  
  ##plot out heatmap
  heatmap.2(avgExpression, trace='n', ...)
}

#' Group density plots
#' 
#' @name plotGroups
#' @aliases plotGroups.formula
#' @aliases plotGroups.deafult
#' @description Generate averaged plots for groups of samples
#' 
#' @param formula      an object of class "\code{\link[stats]{formula}}" (or one that can 
#'                     be coerced to that class): a symbolic description of which fields 
#'                     to use to group the data in tmap.
#' @param data         data frame containing grouping information for samples
#' @param tmap         The tmap object containing a tSNE.map
#' @param grp.info     data frame with two columns, OR a named vector; a description of 
#'                     the sample grouping in tmap. The first column (or the names of the 
#'                     vector) is a list of filenames that match the values in 
#'                     \code{tmap$sourceFile}, and the second column (or values in the 
#'                     vector) is the group membership of that file.
#' @param grps         vector indicating which groups to plot. The groups defined in 
#'                     grp.info will be plotted based on the order in gropus, with NA 
#'                     values being treated as blank plots.
#' @param mfrow        vector of length 2, to be passed on to par. Determines number of 
#'                     rows/columns to include in each plot. If not supplied, an 
#'                     appropriate mfrow will be determined automatically.
#' @param xbins        number of bins that divide the x axis. See hexplot for details
#' @param highlight.clusters vector; a list of clusters to specifically highlight when
#'                     plotting. Clusters not included in this list will be greyed out.
#' @param greyScale    numeric; a value between 0 and 1. If highlight.clusters is defined, 
#'                     this parameter is used to tune how much the non-highlighted 
#'                     clusters are greyed out. 1 is no greying, 0 is completely white. 
#'                     Default is 0.2.
#' @param addEdges     boolean; if true, the cluster edges will be outlined
#' @param edgeCol      the color of the cluster edges.
#' @param lwd          the line width of the cluster edges.
#' @param rasterize    boolean; should the image be rasterized (converted to png before 
#'                     plotting)? Useful to reduce file size when plotting to pdf 
#'                     (or other vector-based device).
#' 
#' @export
plotGroups <- function(x, ...) UseMethod("plotGroups")
#' @rdname plotGroups
#' @export
plotGroups.formula <- function(formula, data, tmap,...){
  data = as.data.frame(data)
  if(!is.data.frame(data)){stop("could not coerce data to data.frame")}
  
  mf <- model.frame(formula, data=data)
  
  grp.info = data.frame(files = mf[,1],
                        grps = do.call(function(...){paste(...,sep="_")}, mf[,-1,drop=F])
                       )
  if(!any(grp.info[,1] %in% tmap$sourceFile)){
    stop("Names in data do not match tmap")
  }
  
  plotGroups.default(tmap, grp.info=grp.info, ...)
}
#' @rdname plotGroups
#' @export
plotGroups.default <- function(tmap, grp.info, grps=NULL, mfrow=NULL,
                               pctile.trim = NULL, pctMax=NULL, 
                               highlight.clusters=NULL, greyScale=0.2,
                               addEdges=!is.null(tmap$binInfo), edgeCol="#00000044",lwd=3,
                               rasterize = (names(dev.cur())=="pdf"), ...){
  
  ##first, check inputs
  if(class(tmap)!="tmap" || is.null(tmap$tSNE.map)){
    stop("tmap must be a valid 'tmap' object")
  }
  
  ##try and coerce grp.info to a data frame
  if(!is.data.frame(grp.info)){
    grp.info = as.data.frame(grp.info)
    if(ncol(grp.info)==1 && !is.null(rownames(grp.info))){
      grp.info = as.data.frame(cbind(sample.name=rownames(grp.info), grp.info))
    }
  }
  if(!is.data.frame(grp.info) || ncol(grp.info) != 2){
    stop("cannot coerce grp.info to 2-column data frame.")
  }
  
  grp.assign = grp.info[,2]
  ugrp = as.character(unique(grp.assign))
  ##if they don't specify an order
  if(is.null(grps)){
    grps = ugrp
  }else{
    grps[!(grps %in% ugrp)] = NA
    if(all(is.na(grps))){stop("group IDs don't match grps")}
  }
  grps = as.character(grps)
  if(length(grps)==0){stop("no groups to plot")}
  
  ##pull out the filenames for each cell
  rns = tmap$sourceFile
  ##we'll compare them against the filenames in the first column of grp.info
  fns = grp.info[,1]
  if(is.null(rns)){stop("sourceFile not defined")}
  
  ##x and y ranges
  xrng = range(tmap$tSNE.map[,1])
  yrng = range(tmap$tSNE.map[,2])
  
  ## Calculate hexbin objects for each file
  hexmaps = lapply(fns, function(n){
    if(is.na(n)){return(NULL)}
    toPlot = which(rns == n)
    ##make sure the group has any cells
    if(length(toPlot)==0){
      warning("file ",n," contains no cells.")
      return(NULL)
    }
    ##make the hexbin object
    return(
      hexplot(tmap$tSNE.map[toPlot,], smooth=T, plot=F,min.grp=1, xlim=xrng,ylim=yrng,...)
    )
  })
  
  ##calculate the total hexbin (for figuring out cluster membership)
  if(!is.null(highlight.clusters)){
    hexmap.full = hexplot(tmap$tSNE.map, smooth=T, plot=F,min.grp=1, xlim=xrng,ylim=yrng,...)
    cell.clust = split(tmap$clustInfo$cluster,hexmap.full@cID)
    cell.clust = sapply(cell.clust, function(x){
      if(length(x)==0){return(NA)}
      names(which.max(table(x)))
    })
  }
  
  ##calc average percents across groups
  hex.max = 0  ##also record percent max at the same time
  hexmaps.avg = lapply(grps,function(n, env=parent.frame(2)){
    if(is.na(n)){return(NULL)}
    ##calculate percents for each file
    pcts = lapply(hexmaps[grp.assign==n], function(hex){
      if(is.null(hex)) return(NA)
      i = hex@cell; nbin = hex@dimen[1]*hex@dimen[2]
      ##pct is a vector containing all possible cells
      pct = rep(0, nbin)
      ##populate those cells that are defined by hex@cell
      pct[i] = hex@count/sum(hex@count);
      return(pct)
    })
    ##average percents
    pcts = do.call(cbind, pcts[!is.na(pcts)])
    pcts = apply(pcts, 1, mean)
    
    ##redefine max pct
    m = max(pcts)
    if(env$hex.max < m){env$hex.max = m}
    
    ##make new "hexbin" -- sorta
    hex = hexmaps[[which(!sapply(hexmaps, is.null))[1]]]
    hex@cell = which(pcts!=0)
    ##hexbins don't like floating point numbers, so converting to integer
    hex@n = as.integer(ceiling(1/min(pcts[hex@cell])))  ##total number of counts
    hex@count = as.integer(ceiling(pcts[hex@cell] * hex@n))
    hex@ncells = length(hex@count)
    return(hex)
  })
  
  ##set max percent
  if(is.null(pctMax)){
    pctMax = hex.max
  }
  
  ###### Generate plots ##########
  ##automatically determine mfrow if not supplied
  if(is.null(mfrow)){
    mfrow=c(1,1)
    l = length(ugrp)
    if(l>3){mfrow[1] = 2}
    if(l>1){mfrow[2] = ifelse(l %in% c(2,4), 2, 3)}
  }
  plotPerPage = mfrow[1]*mfrow[2]
  
  
  ##open new device for rasterizing
  if(rasterize) {
    dev=dev.cur(); ds = dev.size("in"); 
    op = par(mai=c(0,0,0,0), mfrow=c(1,1))
    fn = tempfile()
    png(fn,ds[1],ds[2], units="in", res = 200)
    par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
  }else{
    ##set plotting parameters
    op = par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
  }
  
  ##plot all groups
  for(i in 1:length(grps)){
    main = grps[i]
    hex = hexmaps.avg[[i]]
    
    if(is.null(hex)){plot.new()}##make blank plot if hex is NA
    else{
      if(is.null(highlight.clusters)){
        ##if we're not highlighting clusters, just use default hexplot
        hexplot(hex, norm=T, pctMax=pctMax, axes=F, main=main, min.grp=1, 
                xlim=xrng, ylim=yrng,...)
      }  
      ##if we're highlighting clusters, calculate the colors here
      else{
        ##process the counts
        n = hex@count 
        n[n==0] = NA
        n = n/sum(n, na.rm=T)  ##normalize
        n = log2(n);  ##transform
        
        cols = densityCols(n, norm=T, pctMax=log2(pctMax))
        
        ##find which cells to grey out
        toKeep = names(cell.clust)[which(cell.clust %in% highlight.clusters)]
        toGrey = !(hex@cell %in% toKeep)
        
        ##rather than add an alpha channel to those cols, let's actually make them lighter
        colMat = col2rgb(cols[toGrey])
        colMat = (colMat*greyScale + 255*(1-greyScale))
        cols[toGrey] = rgb(t(colMat),maxColorValue = 255)
        
        hexplot(hex, binCol=cols, axes=F, main=main, min.grp=1, 
                xlim=xrng, ylim=yrng,...)
      }
      
      
      box()
      
      if(addEdges){tmap = plotClusterEdges(tmap, lwd = lwd, line.col=edgeCol)}
    }
    
    ##rasterize
    if(rasterize && i%%plotPerPage==0){
      dev.off()
      plotPNG = readPNG(fn)
      dev.set(dev)
      plot(c(0,1),c(0,1),type="n");rasterImage(plotPNG,0,0,1,1)
      png(fn,ds[1],ds[2], units="in", res = 200)
      par(mfrow=mfrow,mar=c(0.25,0.25,2,0.25))
    }
  }
  ## if rasterize, plot last page
  if(rasterize){
    dev.off();
    dev.set(dev)
    if(i%%plotPerPage!=0){
      plotPNG = readPNG(fn)
      plot(c(0,1),c(0,1),type="n");rasterImage(plotPNG,0,0,1,1)
    }
  }
  
  ##reset plotting parameters back to original
  par(op)
}



############# Helper Functions #################
getBinCenters <- function(hex){
  sx <- hex@xbins/diff(hex@xbnds)
  sy <- (hex@xbins * hex@shape)/diff(hex@ybnds)
  
  nrow = hex@dimen[1]
  ncol = hex@dimen[2]
  nr <- 0:(nrow*ncol-1) %/% (ncol)
  nc <- 0:(nrow*ncol-1) %%  (ncol)
  y.c <- (sqrt(3)/2)/sy * nr + hex@ybnds[1]
  x.c <- 1/sx * ifelse(nr %% 2 == 0, nc, nc + 0.5) + hex@xbnds[1]
  
  return(cbind(x.c, y.c))
}
getBinSize <- function(hex){
  sx <- hex@xbins/diff(hex@xbnds)
  sy <- (hex@xbins * hex@shape)/diff(hex@ybnds)
  inner <- 0.5
  outer <- (2 * inner)/sqrt(3)
  dx <- inner/sx
  dy <- outer/(2 * sy)
  return(c(dx, dy))
}
getBinEdges <- function(dxy){
  ## (point order from hexcoords: ._r_._br_._bl_._l_._tl_.)
  hc = hexcoords(dxy[1],dxy[2])
  with(hc,list(
    l  = list(x=c(x[4],x[5]), y=c(y[4],y[5])),
    bl = list(x=c(x[3],x[4]), y=c(y[3],y[4])),
    br = list(x=c(x[2],x[3]), y=c(y[2],y[3]))
  ))
}