#' Save tSNE data to directory
#' 
#' @name saveTSNEdat
#' @description Saves the matrix and run parameters to "data.dat" and creates 
#'              "dimnames.dat" file containing row and column names 
#' @export
saveTSNEdat<-function(mat, outFolder, theta=0.5, perplexity=30, num_dims=2, 
                      num_proc=NULL, randomSeed=NULL, saveDimnames=TRUE){
  
  ##check if we can save dimnames
  if(saveDimnames && (is.null(rownames(mat)) || is.null(colnames(mat))) ){
    stop("saveDimnames is TRUE, but dimnames not defined!")
  }
  
  ##make sure everything is in the right format
  r = as.integer(nrow(mat))
  c = as.integer(ncol(mat))
  perplexity = as.double(perplexity)
  num_dims = as.integer(num_dims)
  
  ##get number of processors (default is all of them)
  if(is.null(num_proc)){
    num_proc = detectCores()
  }
  num_proc = as.integer(num_proc)
  
  ##get random seed
  if(is.null(randomSeed)){
    randomSeed = .Machine$integer.max*runif(1)
  }
  randomSeed = as.integer(randomSeed)
  
  ##open up connection
  fn = paste0(outFolder,"/data.dat")
  con = file(fn, "wb")
  
  ##write to file
  writeBin(c(r,c), con,size = 4)
  writeBin(c(theta,perplexity), con, size=8)
  writeBin(c(num_dims, num_proc),con,size=4)
  ##write matrix to file
  apply(mat, 1, function(row){writeBin(row, con, size=8)})
  
  writeBin(randomSeed, con, size=4)
  
  ##close connection
  close(con)
  
  ########
  ## save the dimnames
  if(saveDimnames){
    ##encode rownames
    rn.enc = rle(rownames(mat))
    
    ##just to be extra sure there are no commas in the dimnames
    cns = sub(",",".",colnames(mat))
    rn.v = sub(",",".",rn.enc$values)
        
    cns = paste0(cns,collapse=",")
    rn.v = paste0(rn.v,collapse=",")
    rn.l = paste0(rn.enc$lengths,collapse=",")
    ##write to file
    fn = paste0(outFolder, "/dimnames.txt")
    con = file(fn, "wb")
    writeLines(c(cns,rn.v,rn.l),con)
    close(con)
  }
  
  invisible(outFolder)
}

#' Load tSNE data from directory
#' 
#' @name loadTSNEdat
#' @description  Load tSNE data from a directory
#' @details This function assumes that the following files exist in "dir":
#'               - data.dat
#'               - dimnames.txt
#'               - points.txt
#' @export
loadTSNEdat <- function(dir, binaryOutput=T){
  out = list()
  
  ##read in marker vals and run parameters from data.dat
  fn = paste0(dir,"/data.dat")
  if(!file.exists(fn)){stop("cannot find data.dat in specified folder")}
  con = file(fn, "rb")
  
  nrow  = readBin(con, "int",   1,size = 4)
  ncol  = readBin(con, "int",   1,size = 4)
  
  runParams = list()
  runParams$theta      = readBin(con, "double",1,size = 8)
  runParams$perplexity = readBin(con, "double",1,size = 8)
  runParams$num_proc   = readBin(con, "int",   1,size = 4)
  runParams$randomSeed = readBin(con, "int",   1,size = 4)
  out$runParams = runParams
  
  dat = readBin(con, "double",nrow*ncol, size=8)
  out$markerVals = matrix(dat, nrow,ncol,byrow=T)
  rm(dat)
  
  close(con)
  
  ##read in dimnames.txt
  fn = paste0(dir,"/dimnames.txt")
  if(file.exists(fn)){
    con = file(fn, "rb")
    dn = readLines(con)
    close(con)
    
    dn = strsplit(dn,",")
    
    ##put colnames on dat
    if(ncol(out$markerVals)!=length(dn[[1]])){stop("number of columns in data.dat does not match length of colnames.")}
    colnames(out$markerVals) = dn[[1]]
    
    ##make the sourceFile vector
    out$sourceFile = rep(dn[[2]],dn[[3]])  
    
  }else{warning("dimnames.txt not found. Names will not be read.")}
  
  ##if points.txt exists, read in tSNE map from there
  if(binaryOutput){fn = paste0(dir,"/result.dat")}
  else{fn = paste0(dir,"/points.txt")}
  if(file.exists(fn)){
    if(binaryOutput){
      con = file(fn, "rb")
      
      nrow  = readBin(con, "int",   1,size = 4)
      ncol  = readBin(con, "int",   1,size = 4)
      
      dat = readBin(con, "double",nrow*ncol, size=8)
      out$tSNE.map = matrix(dat, nrow,ncol,byrow=T)
      ##out$landmarks = readBin(con, "int", nrow, size=8)
      close(con)
    }else{
      out$tSNE.map = read.csv(fn, header=F)
    }
  }else{warning("points.txt not found. tSNE map will not be read.")}
  
  return(new_tmap(out))
}


#' Function for running tSNE
#' 
#' @description Run the tSNE algorithm on data in a directory
#' 
#' @param dir           The directory containing data to be processed.
#' @param commandOnly   logical; Should the commands for running tSNE from the commandline
#'                      be printed?
#' @param wait          logical; Should R wait for the command to be finished? Ignored if
#'                      \code{commandOnly} is \code{TRUE}.
#'                     
#' @details 
#' This function runs the tSNE algorithm on data saved using the \code{\link{saveTSNEdat}}
#' function. 
#' 
#' @return 
#' (silently), the input directory. Use \code{\link{loadTSNEdat}} to load the results.
#' @export
runTSNE <- function(dir, commandOnly=F, wait=T){
  ##change directory to 
  prevDir = getwd()
  setwd(normalizePath(dir))
  
  ##get tSNE exe location
  tSNE = file.path(system.file("tSNE",package="adjClust"),"pbh_tsne")
  if(!file.exists(tSNE)){stop("tSNE has not been compiled yet! Please run compileTSNE first.")}
  
  ##if they just want the command, print the tSNE exe location
  if(commandOnly){
    cat(paste0("cd ", dir,"\n",tSNE, "\n"))
  }else{
  
    if(.Platform$OS.type=="unix"){
      cmd = paste("script -q /dev/null",tSNE,"| tee tsne.log")
    }else{
      cmd = paste(tSNE, "> tsne.log 2>&1")
    }
    file.remove("tsne.log")
    
    ##run tSNE
    system(cmd, wait=F, intern=F, ignore.stdout = T)
    
    ##track tSNE status
    if(wait){
      pb = NULL;
      msg = "Starting tSNE..."
      
      ##wait til the log file has been created
      while(!file.exists("tsne.log")){Sys.sleep(1)}
      
      ##iterate until the thing is finished
      while(msg!="Finished!"){
        
        status = getTSNEstatus("tsne.log")
        newStep = (status$msg != msg)
        
        ##if we're building the tree
        if(status$msg=="Step 2: Building tree..."){
          if(is.null(pb)){
            cat(status$msg, "\n"); newStep=F
            pb = txtProgressBar(style=3, width=min(getOption("width")-10, 20))
          }
          setTxtProgressBar(pb, status$pct)
        }
        ##if we finished building the tree
        if(newStep & msg=="Step 2: Building tree..."){
          setTxtProgressBar(pb, 1)
          pb=NULL
          cat( status$msg,"\n")
          newStep=F
        }
        
        ##if we're learning the embedding
        if(status$msg=="Step 3: Learning embedding..."){
          if(is.null(pb)){
            pb = txtProgressBar(style=3,width=min(getOption("width")-10, 20))
          }
          setTxtProgressBar(pb, status$pct)
        }
        ##if we finished learning the embedding
        if(newStep & msg=="Step 3: Learning embedding..."){
          setTxtProgressBar(pb, 1)
          pb=NULL
          cat(status$msg,"\n")
          newStep=F
        }
        
        msg = status$msg
        if(newStep){cat(msg,"\n")}
        flush.console()
        Sys.sleep(5)
      }
    }
  }
  
  setwd(prevDir)
  invisible(dir)
}


#' Function to determine the status of the current tSNE run
getTSNEstatus = function(file){
  ##if it hasn't been created yet
  if(!file.exists(file)){return(list(msg=""))}
  lns = readLines(file)
  ##if it hasn't started yet
  if(length(lns)==0){return(list(msg=""))}
  
  ##otherwise, figure out what step we're on
  
  if(any(grepl("Wrote the .* successfully", lns) )){
    return(list(msg = "Finished!"))
  }
  if(length( (i = grep("Learning embedding", lns)) )>0){
    status = "Step 3: Learning embedding..."
    
    ##if it's done
    if(any(grepl("Fitting performed",lns))){
      return(list(msg=status, pct=1))
    }
    
    ##determine percent complete
    ##find total number of iterations
    j = grep("Using \\d+ iterations", lns)
    j.loc = regexpr(" \\d+ ",lns[j])
    iterMax = substr(lns[j],j.loc+1, j.loc+attr(j.loc, "match.length")-2)
    
    ##find current iteration
    iter.loc = regexpr(" \\d+:", lns[length(lns)])
    iter = substr(lns[length(lns)],iter.loc+1, iter.loc+attr(iter.loc, "match.length")-2)
    
    pct = as.numeric(iter)/as.numeric(iterMax)
    
    return(list(msg=status, pct=pct))
  }
  if(length( (i = grep("Building tree", lns)) )>0){
    status = "Step 2: Building tree...";
    
    if(any(grepl("Done in",lns[i:length(lns)]))){
      return(list(msg=status, pct=1))
    }
    
    ##get last entry
    j = length(lns)
    j.loc = gregexpr("\\d+",lns[j])[[1]]
    points = substr(lns[j], j.loc[1], j.loc[1]+attr(j.loc, "match.length")[1]-1)
    total = substr(lns[j], j.loc[2], j.loc[2]+attr(j.loc, "match.length")[2]-1)
    
    pct = as.numeric(points)/as.numeric(total)
    
    return(list(msg=status, pct=pct))
  }
  if(any(grepl("Computing input similarities", lns))){
    return(list(msg = "Step 1: Computing input similarities..."))
  }
  
  return(list(msg=""))
}

#' Compile tSNE
#' @description Function for compiling tSNE
#' @param compiler       string; the c++ compiler to use; defaults to R's default compiler
#' @param use_omp        boolean; should tSNE be compiled with openMP (parallel) support?
#' @param blas_flags     string; the flag used by the C++ compiler for including that blas
#'                       libraries should be included. Default is "-lcblas".
#' @param cblas_loc      string; the location of the cblas.h library. By default, the 
#'                       function will attempt to search the system for a copy of cblas.h.
#'                       If no copy can be found, an error is returned. 
#' @param omp_flag       string; the flag used by the C++ compiler for indicating openMP
#'                       should be used.
#' 
#' @export
compileTSNE <- function(compiler=NULL, use_omp=F, blas_flag="-lcblas", cblas_loc=NULL,
                        omp_flag="-fopenmp"){
  ##get code directory
  codeDir = system.file("tSNE",package="adjClust")
  
  ##change wd to code dir
  oldDir = getwd()
  setwd(codeDir)
  
  ##determine default compiler (the stupid way)
  if(is.null(compiler)){
    cmd = system2("R",args = c("CMD SHLIB -n ptsne.cpp sptree.cpp -o pbh_tsne.so"),
                  stdout = T)
    i = which(cmd=="make would use")+1
    compiler = substr(cmd[i], 1, regexpr(" ",cmd[i])-1)
  }
  
  ##find the location of the cblas.h file
  cat("Finding blas installation...")
  if(.Platform$OS.type=="unix"){
    lns = try(system2("find", args="/ -name 'cblas.h' 2>/dev/null",stdout=T),silent=T)
    if(class(lns)=="try-error"){stop(lns)}
    if(!any(grepl("cblas.h",lns))){stop("Could not find cblas.h library")}
    blasLoc = grep("cblas.h",lns,value=T)[1]
  }else{
    lns = try(system2("dir", args="C:\\cblas.h /s",stdout=T),silent=T)
    if(class(lns)=="try-error"){stop(lns)}
    if(!any(grepl("cblas.h",lns))){stop("Could not find cblas.h library")}
    blasLoc = grep("cblas.h",lns,value=T)[1]
  }
  blasLoc = sub("cblas.h","",blasLoc)
  cat("Done!\n")
  
  ##construct compile command
  cmd = paste0(compiler, " ptsne.cpp sptree.cpp -o pbh_tsne -I.",.Platform$file.sep,
              "resources -I",blasLoc, " ", blas_flag)
  if(use_omp){cmd = paste(cmd, omp_flag)}
  
  ##compile tSNE
  success = system(cmd)
  
  setwd(oldDir)
  if(success==1){stop("Error! Could not compile tSNE.")}
}



