% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hexplot.R
\name{hexplot}
\alias{hexplot}
\alias{hexplot.default}
\alias{hexplot.hexbin}
\alias{hexplot.tmap}
\title{Custom color Hexplot}
\usage{
hexplot(x, ...)

\method{hexplot}{default}(x, y = NULL, xbins = 150, smooth = F,
  plot = T, min.grp = 5, xlim = NULL, ylim = NULL, xlab = NULL,
  ylab = NULL, log = "", ...)

\method{hexplot}{tmap}(x, xlab = NULL, ylab = NULL, ...)

\method{hexplot}{hexbin}(hex, x = NULL, y = NULL, col = NULL,
  binCol = NULL, col.fn = NULL, count.transform = log2,
  pctile.trim = NULL, min.grp = ifelse(is.null(x), 1, 5), cex = 1,
  calc.cex = F, calc.alpha = !is.null(col), alpha.ranges = c(0.2, 0.6),
  smooth = F, norm = F, pctMax = NULL, log = "", xlab = NULL,
  ylab = NULL, add = F, pch = 16, ...)
}
\arguments{
\item{x, y}{the coordinates of points in the plot, or a hexbin object.}

\item{xbins}{the number of bins partitioning the range of x.}

\item{smooth}{boolean; if true, smooth.hexbin is called to smooth densities
of adjacent hexbins}

\item{plot}{logical; if false, plotting will be supressed, and the hexbin
object will be returned instead.}

\item{min.grp}{Minimum number of points to combine together into a bin}

\item{xlim, ylim}{x and y limits of the plotting region.}

\item{xlab, ylab}{x and y axis labes}

\item{log}{character string specifying which axes should be log-scale.}

\item{hex}{a hexbin object}

\item{col}{vector of same length as X describing coloring information
for each point. If values are between 0 and 1, the values are
averaged across all points in the bin, and bin colors are 
calculated using col.fn. Otherwise, values must be valid 
color definitions. Specifying NULL will result in rainbow 
colors being used, as in the main hexplot function.}

\item{col.fn}{a colorRamp function which takes values from 0 to 1 and 
returns RGB specifications. Default is a colorRamp from Green -> Black -> Red.}

\item{count.transform}{function; the transformation to apply to counts before 
calculating colors. Useful for visualizing lower tail of 
distribution.}

\item{pctile.trim}{value defining the maximum percentile for calculating colors.
Only used if col=NULL (see details).}

\item{cex}{cex of largest map points. (see details for information on 
how the cex for each point is calculated).}

\item{calc.cex}{boolean; if true, then cex sizes will be calculated based on
number of points in each bin. Used for displaying density if
custom colors are provided}

\item{calc.alpha}{boolean; if true, then alpha channels will be added to show
density. Use as an alternative to calc.cex.}

\item{alpha.ranges}{numeric vector of length 2; describes the alpha value for the
minimum and median bin size (see details).}

\item{norm}{boolean; if true, convert counts to percentage before 
normalization and calculating colors.}

\item{pctMax}{Alternative to pctile.trim. Specifies the maximum seperable 
value as a percent of the total. Useful for consistency of
color across multiple plots. Only used if norm=T}

\item{add}{logical; if false, plot.new will not be called}

\item{pch, ...}{default plotting parameters to pass to 'plot'}
}
\value{
hexplot.default invisibly returns the generated hexbin object
}
\description{
high density plotting using hexagonal binning with overlaid color
             information.
}
\details{
This function is built off "hexbin" package as an alterative to their less versatile
plotting function. The plot is created by breaking the plotting region up into equally
sized hexagons, counting the number of points inside each, and then using the count to 
determine the color of that bin. 

In order to make the hexbin package interface more cleanly with the default plot
function, the hexagons are represented by single points at the midpoint of each 
hexagonal region rather than as hexagons, like those used in the actual hexbin package.

In some cases, you will have large outlier bins which dominate the color range.
In such cases it is useful to set the max of the color range to a specific percentile
by setting the pctile.trim parameter. This will set all bins in that percentile and
above to the same color, enhancing the separation between bins below that percentile.

If necessary, additional color information can be supplied for each individual 
point, supplanting the default point density colors. This can be done by supplying a 
vector of colors to the \code{col} parameter. 
The \code{col} parameter takes a vector of colors for each individual point, and the 
final plotted color is defined by averaging the RGB values of all points in the bin. 
Alternatively, \code{col} can be a numeric vector with values ranging from 0 to 1, 
representing some continuous property of the map. In this case, the numeric values are
averaged for each point in the bin and are then converted to colors using the colorRamp
function defined by \code{col.fn} The result is a heatmap-like visualization of the 
provided parameter.

In cases where \code{col}, instead of displaying density information using color, hexbin 
densities are displayed using either varying point sizes (if calc.cex is TRUE), or 
using varying levels of transparency (if calc.alpha is TRUE). 
Point sizes and alpha parameters for the plot are calculated in a similar way as the 
colors in hexplot. Bin sizes are first calculated and transformed (using the 
\code{count.transform} function). If \code{calc.cex} is true,
the bin sizes are then scaled such that the average bin has a value of \code{cex/2}. 
Then, all bins greater than \code{cex} are set to \code{cex}.
If \code{calc.alpha} is TRUE, the counts will be scaled such that the minimum value is
equal to \code{alpha.ranges[1]}, and the mean value will be equal to 
\code{alpha.ranges[2]}. Then, all alpha values greater than 1 are set to 1.
}
\examples{
##create dataset
data = rbind(matrix(rnorm(40000), ncol=2),
             matrix(rnorm(40000), ncol=2) + 2,
             matrix(c(rnorm(20000)+2,rnorm(20000)-2), ncol=2)
            )
            
##standard density-colored hexplot
hexplot(data, xbins=100)

##altering the color parameters
hexplot(data, xbins = 100, count.transform=NULL, pctile.trim=95)

##add custom color information
cols = rep(c("red","green","blue"), each=20000)
hexplot(data, xbins=100, col=cols)

}
\author{
Chris Bolen
}

